import pytest

from upsies import utils


@pytest.mark.parametrize(
    argnames='country, exp_name',
    argvalues=(
        ('Peoples Republic of China', 'China'),
        ('tAiWaN', 'Taiwan'),
        ('asdf', 'asdf'),
        (('iceland', 'tv', 'GhA', 'Britain', 'asdf'), ('Iceland', 'Tuvalu', 'Ghana', 'United Kingdom', 'asdf')),
    ),
    ids=lambda v: repr(v),
)
def test_country_name(country, exp_name):
    name = utils.country.name(country)
    assert name == exp_name


@pytest.mark.parametrize(
    argnames='country, exp_code',
    argvalues=(
        ('Britain', 'GB'),
        ('United States', 'US'),
        ('Republic of InDoNeSiA', 'ID'),
        (('iceland', 'tv', 'GhA', 'britain', 'asdf'), ('IS', 'TV', 'GH', 'GB', 'asdf')),
    ),
    ids=lambda v: repr(v),
)
def test_country_code_2letter(country, exp_code):
    code = utils.country.iso2(country)
    assert code == exp_code


@pytest.mark.parametrize(
    argnames='country, exp_tld',
    argvalues=(
        ('United Kingdom', 'uk'),
        ('asdf', 'asdf'),
        (('iceland', 'tv', 'GhA', 'britain', 'asdf'), ('is', 'tv', 'gh', 'uk', 'asdf')),
    ),
    ids=lambda v: repr(v),
)
def test_country_tld(country, exp_tld):
    tld = utils.country.tld(country)
    assert tld == exp_tld



import sys

import pytest

from upsies.utils.imghosts import common


def make_AttributeError(attribute):
    print(sys.version_info, sys.version_info >= (3, 11, 0))
    if sys.version_info >= (3, 11, 0):
        return f"property '{attribute}' of 'UploadedImage' object has no setter"
    elif sys.version_info >= (3, 10, 0):
        return f"can't set attribute '{attribute}'"
    else:
        return "can't set attribute"


def test_UploadedImage_is_URL_string():
    image = common.UploadedImage('http://url.to/image.jpg')
    assert image == 'http://url.to/image.jpg'
    assert isinstance(image, str)
    assert repr(image) == "UploadedImage('http://url.to/image.jpg')"


def test_UploadedImage_thumbnail_url():
    image = common.UploadedImage(
        'http://url.to/image.jpg',
        thumbnail_url='http://url.to/image.thumbnail.jpg',
    )
    assert image.thumbnail_url == 'http://url.to/image.thumbnail.jpg'
    with pytest.raises(AttributeError, match=rf"^{make_AttributeError('thumbnail_url')}$"):
        image.thumbnail_url = 'foo'
    assert repr(image) == "UploadedImage('http://url.to/image.jpg', thumbnail_url='http://url.to/image.thumbnail.jpg')"


def test_UploadedImage_delete_url():
    image = common.UploadedImage(
        'http://url.to/image.jpg',
        delete_url='http://url.to/image.jpg/delete',
    )
    assert image.delete_url == 'http://url.to/image.jpg/delete'
    with pytest.raises(AttributeError, match=rf"^{make_AttributeError('delete_url')}$"):
        image.delete_url = 'foo'
    assert str(image) == 'http://url.to/image.jpg'
    assert repr(image) == "UploadedImage('http://url.to/image.jpg', delete_url='http://url.to/image.jpg/delete')"


def test_UploadedImage_edit_url():
    image = common.UploadedImage(
        'http://url.to/image.jpg',
        edit_url='http://url.to/image.jpg/edit',
    )
    assert image.edit_url == 'http://url.to/image.jpg/edit'
    with pytest.raises(AttributeError, match=rf"^{make_AttributeError('edit_url')}$"):
        image.edit_url = 'foo'
    assert str(image) == 'http://url.to/image.jpg'
    assert repr(image) == "UploadedImage('http://url.to/image.jpg', edit_url='http://url.to/image.jpg/edit')"


@pytest.mark.parametrize(
    argnames='kwargs, exp_repr',
    argvalues=(
        ({'thumbnail_url': 'http://url.to/image.thumbnail.jpg'},
         "UploadedImage('http://url.to/image.jpg', thumbnail_url='http://url.to/image.thumbnail.jpg')"),
        ({'delete_url': 'http://url.to/image.delete.jpg'},
         "UploadedImage('http://url.to/image.jpg', delete_url='http://url.to/image.delete.jpg')"),
        ({'thumbnail_url': 'http://url.to/image.thumbnail.jpg',
         'delete_url': 'http://url.to/image.delete.jpg'},
         ("UploadedImage('http://url.to/image.jpg', "
          "thumbnail_url='http://url.to/image.thumbnail.jpg', "
          "delete_url='http://url.to/image.delete.jpg')")),
    ),
)
def test_UploadedImage_repr(kwargs, exp_repr):
    image = common.UploadedImage('http://url.to/image.jpg', **kwargs)
    assert repr(image) == exp_repr



import pytest

from upsies.utils import timestamp


@pytest.mark.parametrize(
    argnames='seconds, string',
    argvalues=(
        (0, '0:00:00'),
        (1, '0:00:01'),
        (59, '0:00:59'),
        (60, '0:01:00'),
        (61, '0:01:01'),
        (60 * 60 - 1, '0:59:59'),
        (60 * 60, '1:00:00'),
        (10 * 60 * 60 + 1, '10:00:01'),
        (123 * 60 * 60 + 321, '123:05:21'),
    ),
    ids=lambda v: repr(v),
)
def test_pretty_timestamp_from_int(seconds, string):
    assert timestamp.pretty(seconds) == string


@pytest.mark.parametrize(
    argnames='instring, outstring',
    argvalues=(
        ('0', '0:00:00'),
        ('1', '0:00:01'),
        ('123', '0:02:03'),
        ('1:2:3', '1:02:03'),
    ),
    ids=lambda v: repr(v),
)
def test_pretty_timestamp_from_str(instring, outstring):
    assert timestamp.pretty(instring) == outstring

@pytest.mark.parametrize(
    argnames='value, exception, message',
    argvalues=(
        (-1, ValueError, r'^Timestamp must not be negative: -1$'),
        ('-1', ValueError, r'^Timestamp must not be negative: -1$'),
        ('hello', ValueError, r"^Invalid timestamp: 'hello'$"),
        ([1, 2, 3], TypeError, r"^Not a string or number: \[1, 2, 3\]$"),
    ),
    ids=lambda v: repr(v),
)
def test_pretty_timestamp_from_invalid_value(value, exception, message):
    with pytest.raises(exception, match=message):
        timestamp.pretty(value)


@pytest.mark.parametrize(
    argnames='string, seconds',
    argvalues=(
        ('0', 0),
        ('59', 59),
        ('60', 60),
        ('123.2', 123.2),
        ('1:00', 60),
        ('1:01', 61),
        ('1:02:03', 3600 + (2 * 60) + 3),
        ('1:02:03.456', 3600 + (2 * 60) + 3 + 0.456),
        ('1:2:3', 3600 + (2 * 60) + 3),
        ('123:04:55', (123 * 3600) + (4 * 60) + 55),
    ),
    ids=lambda v: repr(v),
)
def test_parse_timestamp_from_str(string, seconds):
    assert timestamp.parse(string) == seconds

@pytest.mark.parametrize(
    argnames='value, exception, message',
    argvalues=(
        ('foo', ValueError, r"^Invalid timestamp: 'foo'$"),
        ('1:00f', ValueError, r"^Invalid timestamp: '1:00f'$"),
        ('-1', ValueError, r'^Timestamp must not be negative: -1$'),
        ('-100', ValueError, r'^Timestamp must not be negative: -100$'),
        ('1:-2:3', ValueError, r'^Timestamp must not be negative: 1:-2:3$'),
    ),
    ids=lambda v: repr(v),
)
def test_parse_timestamp_from_invalid_str(value, exception, message):
    with pytest.raises(exception, match=message):
        timestamp.parse(value)


@pytest.mark.parametrize(
    argnames='value',
    argvalues=(0, 1, 1.5, 59, 60, 61, 300),
    ids=lambda v: repr(v),
)
def test_parse_timestamp_from_number(value):
    assert timestamp.parse(value) == value


@pytest.mark.parametrize(
    argnames='value, exception, message',
    argvalues=(
        (-3, ValueError, r'^Invalid timestamp: -3$'),
        (-61, ValueError, r'^Invalid timestamp: -61$'),
    ),
    ids=lambda v: repr(v),
)
def test_parse_timestamp_from_invalid_number(value, exception, message):
    with pytest.raises(exception, match=message):
        timestamp.parse(value)



import re
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from upsies import errors
from upsies.jobs.dialog import TextFieldJob


@pytest.fixture
def make_TextFieldJob(tmp_path):
    def make_TextFieldJob(**kwargs):
        return TextFieldJob(home_directory=tmp_path, cache_directory=tmp_path, **kwargs)
    return make_TextFieldJob

def test_name_property(make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo')
    assert job.name == 'foo'


def test_label_property(make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo')
    assert job.label == 'Foo'


@pytest.mark.parametrize('read_only', (True, False))
def test_text_property(read_only, make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo', text='0',
                            validator=Mock(side_effect=ValueError('No likey')),
                            normalizer=Mock(side_effect=str.capitalize),
                            read_only=read_only)
    cb = Mock()
    job.signal.register('text', cb.text)
    assert job.text == '0'
    assert cb.text.call_args_list == []
    job.text = 123
    assert job.text == '123'
    assert cb.text.call_args_list == [call('123')]
    job.text = 'foo'
    assert job.text == 'Foo'
    assert cb.text.call_args_list == [call('123'), call('Foo')]


def test_obscured_property(make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo', obscured=True)
    cb = Mock()
    job.signal.register('obscured', cb.obscured)
    assert job.obscured is True
    assert cb.obscured.call_args_list == []
    job.obscured = 0
    assert job.obscured is False
    assert cb.obscured.call_args_list == [call(False)]
    job.obscured = 'yes'
    assert job.obscured is True
    assert cb.obscured.call_args_list == [call(False), call(True)]


def test_read_only_property(make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo', read_only=True)
    cb = Mock()
    job.signal.register('read_only', cb.read_only)
    assert job.read_only is True
    assert cb.read_only.call_args_list == []
    job.read_only = ''
    assert job.read_only is False
    assert cb.read_only.call_args_list == [call(False)]
    job.read_only = '1'
    assert job.read_only is True
    assert cb.read_only.call_args_list == [call(False), call(True)]


def test_is_loading_property(make_TextFieldJob):
    job = make_TextFieldJob(name='foo', label='Foo')
    cb = Mock()
    job.signal.register('is_loading', cb.is_loading)
    assert job.is_loading is False
    assert cb.is_loading.call_args_list == []
    job.is_loading = 1
    assert job.is_loading is True
    assert cb.is_loading.call_args_list == [call(True)]
    job.is_loading = ''
    assert job.is_loading is False
    assert cb.is_loading.call_args_list == [call(True), call(False)]


@pytest.mark.asyncio
async def test_execute_gets_awaitable(make_TextFieldJob, mocker):
    async def text():
        return 'my text'

    awaitable = text()
    default = 'my default'
    finish_on_success = 'my finish_on_success'
    nonfatal_exceptions = 'my nonfatal_exceptions'
    try:

        job = make_TextFieldJob(
            name='foo',
            label='Foo',
            text=awaitable,
            default=default,
            finish_on_success=finish_on_success,
            nonfatal_exceptions=nonfatal_exceptions,
        )

        mocks = Mock()
        mocks.attach_mock(mocker.patch.object(job, 'fetch_text', Mock(return_value='fetch_text() coro')), 'fetch_text')
        mocks.attach_mock(mocker.patch.object(job, 'attach_task'), 'attach_task')
        mocks.attach_mock(mocker.patch.object(job, 'set_text'), 'set_text')

        job.execute()

        assert mocks.mock_calls == [
            call.fetch_text(
                coro=awaitable,
                default='my default',
                finish_on_success='my finish_on_success',
                nonfatal_exceptions='my nonfatal_exceptions',
            ),
            call.attach_task(job.fetch_text.return_value),
        ]

    finally:
        # Prevent RuntimeWarning: coroutine 'test_execute_gets_awaitable.<locals>.text' was never awaited
        await awaitable

@pytest.mark.asyncio
async def test_execute_gets_coroutine_function(make_TextFieldJob, mocker):
    text = Mock()
    default = 'my default'
    finish_on_success = 'my finish_on_success'
    nonfatal_exceptions = 'my nonfatal_exceptions'

    job = make_TextFieldJob(
        name='foo',
        label='Foo',
        text=text,
        default=default,
        finish_on_success=finish_on_success,
        nonfatal_exceptions=nonfatal_exceptions,
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'fetch_text', Mock(return_value='fetch_text() coro')), 'fetch_text')
    mocks.attach_mock(mocker.patch.object(job, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(job, 'set_text'), 'set_text')

    with patch('inspect.iscoroutinefunction', return_value=True):
        job.execute()

    assert mocks.mock_calls == [
        call.fetch_text(
            coro=text.return_value,
            default=default,
            finish_on_success=finish_on_success,
            nonfatal_exceptions=nonfatal_exceptions,
        ),
        call.attach_task(job.fetch_text.return_value),
    ]

@pytest.mark.asyncio
async def test_execute_gets_text(make_TextFieldJob, mocker):
    text = 'my text'

    job = make_TextFieldJob(
        name='foo',
        label='Foo',
        text=text,
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'fetch_text', Mock(return_value='fetch_text() coro')), 'fetch_text')
    mocks.attach_mock(mocker.patch.object(job, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(job, 'set_text'), 'set_text')

    job.execute()

    assert mocks.mock_calls == [
        call.set_text(
            text=text,
            default=None,
            finish_on_success=False,
            nonfatal_exceptions=(),
        ),
    ]

@pytest.mark.asyncio
async def test_execute_gets_anything_else(make_TextFieldJob, mocker):
    text = Mock()
    default = 'my default'
    finish_on_success = 'my finish_on_success'
    nonfatal_exceptions = 'my nonfatal_exceptions'

    job = make_TextFieldJob(
        name='foo',
        label='Foo',
        text=text,
        default=default,
        finish_on_success=finish_on_success,
        nonfatal_exceptions=nonfatal_exceptions,
    )

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'fetch_text', Mock(return_value='fetch_text() coro')), 'fetch_text')
    mocks.attach_mock(mocker.patch.object(job, 'attach_task'), 'attach_task')
    mocks.attach_mock(mocker.patch.object(job, 'set_text'), 'set_text')

    job.execute()

    assert mocks.mock_calls == [
        call.set_text(
            text=text,
            default=default,
            finish_on_success=finish_on_success,
            nonfatal_exceptions=nonfatal_exceptions,
        ),
    ]


@pytest.mark.parametrize(
    argnames='text, default, finish_on_success, nonfatal_exceptions, exp_mock_calls',
    argvalues=(
        (
            'my text', 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call._set_text('my text', default='my default', finish='my finish_on_success'),
            ],
        ),
        (
            Mock(return_value='my text'), 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call._set_text('my text', default='my default', finish='my finish_on_success'),
            ],
        ),
        (
            Mock(side_effect=errors.RequestError('foo')), 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call._set_text('my default'),
                call._handle_exception(errors.RequestError('foo'), nonfatal='my nonfatal_exceptions'),
            ],
        ),
        (
            None, 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call._set_text(None, default='my default', finish='my finish_on_success'),
            ],
        ),
    ),
    ids=lambda v: repr(v),
)
def test_set_text(text, default, finish_on_success, nonfatal_exceptions,
                  exp_mock_calls,
                  make_TextFieldJob, mocker):
    job = make_TextFieldJob(name='foo', label='Foo')

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, '_set_text'), '_set_text')
    mocks.attach_mock(mocker.patch.object(job, '_handle_exception'), '_handle_exception')

    def check_states(*args, **kwargs):
        assert job.is_loading is True
        assert job.read_only is True

    mocks._set_text.side_effect = check_states
    mocks._handle_exception.side_effect = check_states

    job.set_text(
        text,
        default=default,
        finish_on_success=finish_on_success,
        nonfatal_exceptions=nonfatal_exceptions,
    )
    assert mocks.mock_calls == exp_mock_calls
    assert job.is_loading is False
    assert job.read_only is False


@pytest.mark.parametrize(
    argnames='coro, default, finish_on_success, nonfatal_exceptions, exp_mock_calls',
    argvalues=(
        (
            AsyncMock(return_value='my text'), 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call.coro(),
                call._set_text('my text', default='my default', finish='my finish_on_success'),
            ],
        ),
        (
            AsyncMock(side_effect=errors.RequestError('foo')), 'my default', 'my finish_on_success', 'my nonfatal_exceptions',
            [
                call.coro(),
                call._set_text('my default'),
                call._handle_exception(errors.RequestError('foo'), nonfatal='my nonfatal_exceptions'),
            ],
        ),

    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_fetch_text(coro, default, finish_on_success, nonfatal_exceptions,
                          exp_mock_calls,
                          make_TextFieldJob, mocker):
    job = make_TextFieldJob(name='foo', label='Foo')

    mocks = Mock()
    mocks.attach_mock(coro, 'coro')
    mocks.attach_mock(mocker.patch.object(job, 'send'), 'send')
    mocks.attach_mock(mocker.patch.object(job, '_set_text'), '_set_text')
    mocks.attach_mock(mocker.patch.object(job, '_handle_exception'), '_handle_exception')

    async def wrapped_coro():
        assert job.is_loading is True
        assert job.read_only is True
        return await coro()

    await job.fetch_text(
        wrapped_coro(),
        default=default,
        finish_on_success=finish_on_success,
        nonfatal_exceptions=nonfatal_exceptions,
    )
    assert mocks.mock_calls == exp_mock_calls
    assert job.is_loading is False
    assert job.read_only is False


@pytest.mark.parametrize(
    argnames='text, default, finish, exp_text, exp_send_calls',
    argvalues=(
        (None, None, False, 'original', []),
        (None, None, True, 'original', []),
        (None, 'mydefault', False, 'mydefault', []),
        (None, 'mydefault', True, 'mydefault', []),
        ('mytext', None, False, 'mytext', []),
        ('mytext', None, True, 'mytext', [call('mytext')]),
        ('mytext', 'mydefault', False, 'mytext', []),
        ('mytext', 'mydefault', True, 'mytext', [call('mytext')]),
    ),
    ids=lambda v: repr(v),
)
def test__set_text(text, default, finish, exp_text, exp_send_calls, make_TextFieldJob, mocker):
    job = make_TextFieldJob(name='foo', label='Foo')
    job.text = 'original'
    mocker.patch.object(job, 'send')

    job._set_text(text, default=default, finish=finish)
    assert job.text == exp_text

    assert job.send.call_args_list == exp_send_calls


@pytest.mark.parametrize(
    argnames='exception, nonfatal, exp_warn, exp_raised',
    argvalues=(
        (errors.RequestError('foo'), (), None, errors.RequestError('foo')),
        (errors.RequestError('foo'), (errors.RequestError,), errors.RequestError('foo'), None),
        (errors.ConfigError('foo'), (errors.RequestError,), None, errors.ConfigError('foo')),
    ),
    ids=lambda v: repr(v),
)
def test__handle_exception(exception, nonfatal, exp_warn, exp_raised, make_TextFieldJob, mocker):
    job = make_TextFieldJob(name='foo', label='Foo')
    mocker.patch.object(job, 'warn')

    if exp_raised:
        with pytest.raises(type(exp_raised), match=rf'^{re.escape(str(exp_raised))}$'):
            job._handle_exception(exception, nonfatal=nonfatal)
    else:
        job._handle_exception(exception, nonfatal=nonfatal)
        assert job.warn.call_args_list == [call(exp_warn)]


def test_send_valid_text(make_TextFieldJob):
    callables = Mock()
    callables.normalizer = Mock(side_effect=str.upper)
    job = make_TextFieldJob(
        name='foo',
        label='Foo',
        text='bar',
        validator=callables.validator,
        normalizer=callables.normalizer,
    )
    assert callables.mock_calls == [
        call.normalizer('bar'),
    ]
    assert job.text == 'BAR'
    assert job.output == ()

    callables.reset_mock()
    job.send('baz')

    assert callables.mock_calls == [
        call.normalizer('baz'),
        call.validator('BAZ'),
    ]
    assert job.warnings == ()
    assert job.is_finished
    assert job.exit_code == 0
    assert job.output == ('BAZ',)

def test_send_invalid_text(make_TextFieldJob):
    callables = Mock()
    callables.validator = Mock(side_effect=ValueError('Nope'))
    callables.normalizer = Mock(side_effect=str.upper)
    job = make_TextFieldJob(
        name='foo',
        label='Foo',
        text='bar',
        validator=callables.validator,
        normalizer=callables.normalizer,
    )
    assert callables.mock_calls == [
        call.normalizer('bar'),
    ]

    callables.reset_mock()
    job.send('baz')

    assert callables.mock_calls == [
        call.normalizer('baz'),
        call.validator('BAZ'),
    ]
    assert job.warnings == ('Nope',)
    assert not job.is_finished
    assert job.exit_code is None
    assert job.output == ()



import os
import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies.utils import release, webdbs
from upsies.utils.types import ReleaseType


def test_Query_title():
    q = webdbs.Query()
    cb = Mock()
    q.signal.register('changed', cb)

    for _ in range(3):
        assert q.title == ''
        assert q.title_normalized == ''
        assert cb.mock_calls == []

    for _ in range(3):
        q.title = 'The Title '
        assert q.title_normalized == 'the title'
        assert cb.mock_calls == [call(q)]

    for _ in range(3):
        q.title = '  The\n OTHER  Title! '
        assert q.title_normalized == 'the other title!'
        assert cb.mock_calls == [call(q), call(q)]


def test_Query_year():
    q = webdbs.Query('The Title', year='2000')
    cb = Mock()
    q.signal.register('changed', cb)
    assert q.year == '2000'
    assert cb.mock_calls == []

    for _ in range(3):
        q.year = 2001
        assert q.year == '2001'
        assert cb.mock_calls == [call(q)]

    for _ in range(3):
        q.year = 2001.5
        assert q.year == '2001'
        assert cb.mock_calls == [call(q)]

    with pytest.raises(ValueError, match=r'^Invalid year: 1000$'):
        q.year = 1000
    assert cb.mock_calls == [call(q)]
    with pytest.raises(ValueError, match=r'^Invalid year: 3000$'):
        q.year = '3000'
    assert cb.mock_calls == [call(q)]
    with pytest.raises(ValueError, match=r'^Invalid year: foo$'):
        q.year = 'foo'
    assert cb.mock_calls == [call(q)]

    q.year = ''
    assert q.year is None
    assert cb.mock_calls == [call(q), call(q)]


@pytest.mark.parametrize('typ', list(ReleaseType) + [str(t) for t in ReleaseType], ids=lambda v: repr(v))
def test_Query_valid_type(typ):
    q = webdbs.Query('The Title', type=typ)
    cb = Mock()
    q.signal.register('changed', cb)
    assert q.type is ReleaseType(typ)

    all_types = list(ReleaseType)
    if all_types.index(ReleaseType(typ)) < len(all_types) - 1:
        next_type_index = all_types.index(ReleaseType(typ)) + 1
    else:
        next_type_index = 0
    print(typ, all_types, next_type_index)
    next_type = all_types[next_type_index]

    for _ in range(3):
        q.type = next_type
        assert q.type is ReleaseType(next_type)
        assert cb.mock_calls == [call(q)]

    for _ in range(3):
        q.type = typ
        assert q.type is ReleaseType(typ)
        assert cb.mock_calls == [call(q), call(q)]

def test_Query_invalid_type():
    q = webdbs.Query('The Title')
    cb = Mock()
    q.signal.register('changed', cb)
    with pytest.raises(ValueError, match=r"^'foo' is not a valid ReleaseType$"):
        q.type = 'foo'
    assert cb.mock_calls == []


def test_Query_id():
    q = webdbs.Query('The Title', id='12345')
    cb = Mock()
    q.signal.register('changed', cb)
    assert q.id == '12345'
    assert cb.mock_calls == []

    for _ in range(3):
        q.id = 123
        assert q.id == '123'
        assert cb.mock_calls == [call(q)]

    for _ in range(3):
        q.id = 'tt12345'
        assert q.id == 'tt12345'
        assert cb.mock_calls == [call(q), call(q)]


@pytest.mark.parametrize('silent', (False, True))
def test_Query_update(silent):
    q = webdbs.Query('The Title', type=ReleaseType.series)
    cb = Mock()
    q.signal.register('changed', cb)
    assert cb.call_args_list == []

    for _ in range(3):
        q.update(webdbs.Query('The Title', type=ReleaseType.movie), silent=silent)
        assert str(q) == 'The Title type:movie'
        assert cb.call_args_list == [] if silent else [call(q)]

    for _ in range(3):
        q.update(webdbs.Query('The Title', type=ReleaseType.season), silent=silent)
        assert str(q) == 'The Title type:season'
        assert cb.call_args_list == [] if silent else [call(q), call(q)]

    for _ in range(3):
        q.update(webdbs.Query('The Other Title', type=ReleaseType.episode, year='2015'), silent=silent)
        assert str(q) == 'The Other Title type:episode year:2015'
        assert cb.call_args_list == [] if silent else [call(q), call(q), call(q)]

    for _ in range(3):
        q.update(webdbs.Query('The Title', id='123', year='2015', type=ReleaseType.episode), silent=silent)
        assert str(q) == 'id:123'
        assert cb.call_args_list == [] if silent else [call(q), call(q), call(q), call(q)]


def test_Query_copy():
    q = webdbs.Query('The Title', year='2010', type=ReleaseType.series, id='1234')
    cb = Mock()
    q.signal.register('changed', cb)
    assert cb.call_args_list == []

    assert q.copy() == webdbs.Query(
        title='The Title',
        year='2010',
        type=ReleaseType.series,
        id='1234',
    )

    assert q.copy(title='Foo') == webdbs.Query(
        title='Foo',
        year='2010',
        type=ReleaseType.series,
        id='1234',
    )

    assert q.copy(year='2020') == webdbs.Query(
        title='The Title',
        year='2020',
        type=ReleaseType.series,
        id='1234',
    )

    assert q.copy(type=ReleaseType.movie) == webdbs.Query(
        title='The Title',
        year='2010',
        type=ReleaseType.movie,
        id='1234',
    )

    assert q.copy(id='4321') == webdbs.Query(
        title='The Title',
        year='2010',
        type=ReleaseType.series,
        id='4321',
    )

    assert cb.call_args_list == []

@pytest.mark.parametrize(
    argnames=('a', 'b'),
    argvalues=(
        (webdbs.Query('The Title'), webdbs.Query('The Title')),
        (webdbs.Query('The Title'), webdbs.Query('the title')),
        (webdbs.Query('The Title'), webdbs.Query('the title ')),
        (webdbs.Query('The Title'), webdbs.Query(' the title')),
        (webdbs.Query('The Title'), webdbs.Query(' the title ')),
        (webdbs.Query('The Title'), webdbs.Query('the  title')),
        (webdbs.Query('The Title'), webdbs.Query('the  title ')),
        (webdbs.Query('The Title'), webdbs.Query(' the title')),
        (webdbs.Query('The Title', year='2000'), webdbs.Query('The Title', year='2000')),
        (webdbs.Query('The Title', type=ReleaseType.movie), webdbs.Query('The Title', type=ReleaseType.movie)),
        (webdbs.Query('The Title', type=ReleaseType.series), webdbs.Query('The Title', type=ReleaseType.series)),
        (webdbs.Query('The Title', type=ReleaseType.episode, year=2000), webdbs.Query('The Title', type=ReleaseType.episode, year=2000)),
    ),
    ids=lambda value: str(value),
)
def test_Query_equality(a, b):
    assert a == b
    assert b == a

@pytest.mark.parametrize(
    argnames=('a', 'b'),
    argvalues=(
        (webdbs.Query('The Title'), webdbs.Query('The Title 2')),
        (webdbs.Query(id=123), webdbs.Query(id=124)),
        (webdbs.Query('The Title', year='2000'), webdbs.Query('The Title', year='2001')),
        (webdbs.Query('The Title', type=ReleaseType.movie), webdbs.Query('The Title', type=ReleaseType.series)),
        (webdbs.Query('The Title', type=ReleaseType.series, year=2000), webdbs.Query('The Title', type=ReleaseType.movie, year=2000)),
        (webdbs.Query('The Title', type=ReleaseType.series, year=2000), webdbs.Query('The Title', type=ReleaseType.series, year=2001)),
        (webdbs.Query('The Title', type=ReleaseType.series, year=2000), webdbs.Query('The Title', type=ReleaseType.movie, year=2001)),
    ),
    ids=lambda value: str(value),
)
def test_Query_inequality(a, b):
    assert a != b
    assert b != a

@pytest.mark.parametrize(
    argnames=('query', 'exp_string'),
    argvalues=(
        (webdbs.Query('The Title'), 'The Title'),
        (webdbs.Query('The Title', type=ReleaseType.movie), 'The Title type:movie'),
        (webdbs.Query('The Title', type=ReleaseType.series), 'The Title type:season'),
        (webdbs.Query('The Title', type=ReleaseType.movie, year='2010'), 'The Title type:movie year:2010'),
        (webdbs.Query('The Title', id='123', year='2010', type=ReleaseType.movie), 'id:123'),
    ),
    ids=lambda value: str(value),
)
def test_Query_as_string(query, exp_string):
    assert str(query) == exp_string

@pytest.mark.parametrize(
    argnames=('string', 'exp_result'),
    argvalues=(
        ('The Title', webdbs.Query('The Title')),
        ('The Title type:movie', webdbs.Query('The Title', type=ReleaseType.movie)),
        ('The Title type:season', webdbs.Query('The Title', type=ReleaseType.season)),
        ('The Title type:episode', webdbs.Query('The Title', type=ReleaseType.episode)),
        ('The Title type:series', webdbs.Query('The Title', type=ReleaseType.series)),
        ('The Title type:film', webdbs.Query('The Title', type=ReleaseType.movie)),
        ('The Title type:tv', webdbs.Query('The Title', type=ReleaseType.series)),
        ('The Title type:show', webdbs.Query('The Title', type=ReleaseType.series)),
        ('The Title type:tvshow', webdbs.Query('The Title', type=ReleaseType.series)),
        ('The Title type:foo', ValueError('Invalid type: foo')),
        ('The Title year:2000', webdbs.Query('The Title', year='2000')),
        ('The Title year:2003', webdbs.Query('The Title', year='2003')),
        ('The Title year:6000', ValueError('Invalid year: 6000')),
        ('The Title year:123', ValueError('Invalid year: 123')),
        ('The Title year:foo', ValueError('Invalid year: foo')),
        ('The Title type:movie year:2000', webdbs.Query('The Title', type=ReleaseType.movie, year='2000')),
        ('The Title year:2003 type:series', webdbs.Query('The Title', type=ReleaseType.series, year='2003')),
        ('The Title id:foo', webdbs.Query('The Title', id='foo')),
        ('id:foo', webdbs.Query(id='foo')),
        ('id:foo year:2005', webdbs.Query(id='foo', year='2005')),
        ('The Title Id:foo Type:bar Year:baz', webdbs.Query('The Title Id:foo Type:bar Year:baz')),
    ),
    ids=lambda v: str(v),
)
def test_Query_from_string(string, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            webdbs.Query.from_string(string)
    else:
        assert webdbs.Query.from_string(string) == exp_result


@pytest.mark.parametrize('cls', (release.ReleaseInfo, release.ReleaseName))
@pytest.mark.parametrize(
    argnames=('release', 'exp_query'),
    argvalues=(
        ('The Title', webdbs.Query('The Title', type=ReleaseType.movie)),
        ('The Title S03', webdbs.Query('The Title', type=ReleaseType.season)),
        ('The Title S03', webdbs.Query('The Title', type=ReleaseType.season)),
        ('The Title S03E04', webdbs.Query('The Title', type=ReleaseType.episode)),
        ('The Title 2000', webdbs.Query('The Title', year='2000', type=ReleaseType.movie)),
        ('The Title 2001 S03E04', webdbs.Query('The Title', year='2001', type=ReleaseType.episode)),

    ),
)
def test_Query_from_release(cls, release, exp_query):
    obj = cls(release)
    assert webdbs.Query.from_release(obj) == exp_query


@pytest.mark.parametrize(
    argnames=('path', 'exp_ReleaseInfo_call'),
    argvalues=(
        ('path/to/Release Name', call('path/to/Release Name')),
        (123, call('123')),
    ),
    ids=lambda v: repr(v),
)
def test_Query_from_path(path, exp_ReleaseInfo_call, mocker):
    ReleaseInfo_mock = mocker.patch('upsies.utils.webdbs.common.release.ReleaseInfo')
    from_release_mock = mocker.patch('upsies.utils.webdbs.common.Query.from_release')
    return_value = webdbs.Query.from_path(path)
    assert ReleaseInfo_mock.call_args_list == [exp_ReleaseInfo_call]
    assert from_release_mock.call_args_list == [call(ReleaseInfo_mock.return_value)]
    assert return_value == from_release_mock.return_value


@pytest.mark.parametrize(
    argnames=('obj', 'path_exists', 'exp_from_method', 'exp_exception'),
    argvalues=(
        (webdbs.Query('Release Name'), False, None, None),
        ('Release Name', False, 'from_string', None),
        (os.path.join('path', 'to', 'Release Name'), False, 'from_path', None),
        ('path.in.current.directory', True, 'from_path', None),
        ('path.in.current.directory', False, 'from_string', None),
        (release.ReleaseInfo('Release Name'), False, 'from_release', None),
        (release.ReleaseName('Release Name'), False, 'from_release', None),
        ([1, 2, 3], False, None, TypeError('Unsupported type: list: [1, 2, 3]')),
    ),
    ids=lambda v: repr(v),
)
def test_Query_from_any(obj, path_exists, exp_from_method, exp_exception, mocker):
    from_methods = {
        'from_string': mocker.patch('upsies.utils.webdbs.common.Query.from_string'),
        'from_release': mocker.patch('upsies.utils.webdbs.common.Query.from_release'),
        'from_path': mocker.patch('upsies.utils.webdbs.common.Query.from_path'),
    }
    mocker.patch('os.path.exists', return_value=path_exists)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            webdbs.Query.from_any(obj)
    else:
        return_value = webdbs.Query.from_any(obj)
        if exp_from_method:
            assert from_methods[exp_from_method].call_args_list == [call(obj)]
            assert return_value is from_methods[exp_from_method].return_value
        else:
            assert return_value is obj


@pytest.mark.parametrize(
    argnames=('query', 'exp_repr'),
    argvalues=(
        (webdbs.Query('The Title'),
         "Query(title='The Title')"),
        (webdbs.Query('The Title', type=ReleaseType.movie),
         "Query(title='The Title', type=ReleaseType.movie)"),
        (webdbs.Query('The Title', year='1999'),
         "Query(title='The Title', year='1999')"),
        (webdbs.Query('The Title', type=ReleaseType.season, year='1999'),
         "Query(title='The Title', year='1999', type=ReleaseType.season)"),
        (webdbs.Query('The Title', type=ReleaseType.season, year='1999', id='123'),
         "Query(title='The Title', year='1999', type=ReleaseType.season, id='123')"),
    ),
    ids=lambda value: str(value),
)
def test_Query_as_repr(query, exp_repr):
    assert repr(query) == exp_repr


@pytest.mark.asyncio
async def test_SearchResult_with_all_info():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'type' : ReleaseType.movie,
        'year' : '2000',
        'title' : 'Foo',
        'title_original' : 'Le Foo',
        'title_english' : 'Fucking Foo',
        'genres' : ('foo', 'some bar'),
        'cast' : ('This Guy', 'That Dude'),
        'summary' : 'I dunno.',
        'countries' : ('Antarctica',),
    }
    sr = webdbs.SearchResult(**info)
    async_attributes = (
        'cast',
        'countries',
        'directors',
        'genres',
        'summary',
        'title_english',
        'title_original',
    )
    for k, v in info.items():
        if k in async_attributes:
            assert await getattr(sr, k)() == v
        else:
            assert getattr(sr, k) == v

def test_SearchResult_with_only_mandatory_info():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'type' : ReleaseType.series,
        'year' : '2000',
        'title' : 'Foo',
    }
    sr = webdbs.SearchResult(**info)
    for k, v in info.items():
        assert getattr(sr, k) == v

def test_SearchResult_with_lacking_mandatory_info():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'type' : ReleaseType.season,
        'year' : '2000',
        'title' : 'Foo',
    }
    for k in info:
        info_ = info.copy()
        del info_[k]
        with pytest.raises(TypeError, match=rf'missing 1 required keyword-only argument: {k!r}'):
            webdbs.SearchResult(**info_)

def test_SearchResult_with_valid_type():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'year' : '2000',
        'title' : 'Foo',
    }
    for type in tuple(ReleaseType) + tuple(str(t) for t in ReleaseType):
        sr = webdbs.SearchResult(type=type, **info)
        assert sr.type == ReleaseType(type)

def test_SearchResult_with_invalid_type():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'year' : '2000',
        'title' : 'Foo',
    }
    for type in ('foo', 'bar', 'baz', None):
        with pytest.raises(ValueError, match=rf'{type!r} is not a valid ReleaseType'):
            webdbs.SearchResult(type=type, **info)

def test_SearchResult_with_invalid_attribute():
    info = {
        'id' : '123',
        'url' : 'http://foo.bar/123',
        'type' : ReleaseType.series,
        'year' : '2000',
        'title' : 'Foo',
    }
    sr = webdbs.SearchResult(**info)
    with pytest.raises(AttributeError, match=r'foo'):
        sr.foo

@pytest.mark.parametrize(
    argnames=('id', 'exp_type'),
    argvalues=(
        ('123', str),
        (123.0, float),
        (123, int),
    ),
)
def test_SearchResult_preserves_id_type(id, exp_type):
    info = {
        'id' : id,
        'type': ReleaseType.series,
        'url' : 'http://foo.bar/123',
        'year' : '2000',
        'title' : 'Foo',
    }
    result = webdbs.SearchResult(**info)
    assert result.id == id
    assert isinstance(result.id, exp_type)

@pytest.mark.parametrize(
    argnames=('countries', 'exp_countries'),
    argvalues=(
        (['US'], ('United States',)),
        (AsyncMock(return_value=['US']), ('United States',)),
    ),
)
@pytest.mark.asyncio
async def test_SearchResult_normalizes_countries(countries, exp_countries):
    info = {
        'id' : '123',
        'type': ReleaseType.series,
        'url' : 'http://foo.bar/123',
        'year' : '2000',
        'title' : 'Foo',
        'countries': countries,
    }
    result = webdbs.SearchResult(**info)
    countries = await result.countries()
    assert countries == exp_countries

def test_SearchResult_repr():
    info = {
        'id' : '123',
        'title' : 'Foo',
        'type': ReleaseType.movie,
        'url' : 'http://foo.bar/123',
        'year' : '2000',
    }
    result = webdbs.SearchResult(**info)
    exp_repr_regex = re.compile(
        r"^SearchResult\("
        r"id='123', "
        r"title='Foo', "
        r"type=ReleaseType.movie, "
        r"url='http://foo.bar/123', "
        r"year='2000', "
        r"cast=<function async_get_cast at 0x[0-9a-f]+>, "
        r"countries=<function async_get_countries at 0x[0-9a-f]+>, "
        r"directors=<function async_get_directors at 0x[0-9a-f]+>, "
        r"genres=<function async_get_genres at 0x[0-9a-f]+>, "
        r"summary=<function async_get_summary at 0x[0-9a-f]+>, "
        r"title_english=<function async_get_title_english at 0x[0-9a-f]+>, "
        r"title_original=<function async_get_title_original at 0x[0-9a-f]+>"
        r"\)$"
    )
    repr_string = repr(result)
    print(repr_string)
    print(exp_repr_regex.pattern)
    assert exp_repr_regex.search(repr_string)


def test_Person():
    assert isinstance(webdbs.Person('Foo Bar'), str)
    assert webdbs.Person('Foo Bar').url == ''
    assert webdbs.Person('Foo Bar', url='http://foo').url == 'http://foo'

@pytest.mark.parametrize(
    argnames='args, kwargs, exp_repr',
    argvalues=(
        (('Hey Ho',), {}, "Person('Hey Ho')"),
        (('Hey Ho',), {'url': 'http://localhost/hey'}, "Person('Hey Ho', url='http://localhost/hey')"),
        (('Hey Ho', 'http://localhost/hey'), {}, "Person('Hey Ho', url='http://localhost/hey')"),
    ),
)
def test_Person_repr(args, kwargs, exp_repr):
    assert repr(webdbs.Person(*args, **kwargs)), exp_repr



"""
Concrete :class:`~.base.TrackerBase` subclass for NBL
"""

from ... import errors
from ...utils import http
from .. import base
from .config import NblTrackerConfig
from .jobs import NblTrackerJobs

import logging  # isort:skip
_log = logging.getLogger(__name__)


class NblTracker(base.TrackerBase):
    name = 'nbl'
    label = 'NBL'

    setup_howto_template = (
        '{howto.introduction}\n'
        '\n'
        '{howto.current_section}. Announce URL\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to Shows -> Upload and copy the ANNOUNCE_URL.\n'
        '   {howto.current_section}.2 $ upsies set trackers.nbl.announce_url ANNOUNCE_URL\n'
        '{howto.bump_section}'
        '\n'
        '{howto.current_section}. API key\n'
        '\n'
        '   {howto.current_section}.1 On the website, go to USERNAME -> Settings and scroll down\n'
        '       to "API keys".\n'
        '   {howto.current_section}.2 Tick the "New Key" and the "Upload" boxes.\n'
        '   {howto.current_section}.3 Click on "Save Profile".\n'
        '   {howto.current_section}.4 Scroll down to "API keys" again and copy the new API_KEY.\n'
        '   {howto.current_section}.5 $ upsies set trackers.nbl.apikey API_KEY\n'
        '{howto.bump_section}'
        '\n'
        '{howto.autoseed}\n'
        '\n'
        '{howto.upload}\n'
    )

    TrackerConfig = NblTrackerConfig
    TrackerJobs = NblTrackerJobs

    async def _login(self):
        pass

    async def _logout(self):
        pass

    async def get_announce_url(self):
        return self.options['announce_url']

    def get_upload_url(self):
        return self.options['upload_url']

    async def upload(self, tracker_jobs):
        if not self.options['apikey']:
            raise errors.RequestError('No API key configured')

        _log.debug('Uploading to %r', self.get_upload_url())

        # Remove None and empty strings from post_data but keep any int(0)
        # values. Also ensure all keys and values are strings.
        post_data = {
            str(k): str(v)
            for k, v in tracker_jobs.post_data.items()
            if v not in (None, '')
        }
        _log.debug('POST data: %r', post_data)

        files = {
            'file_input': {
                'file': tracker_jobs.torrent_filepath,
                'mimetype': 'application/x-bittorrent',
            },
        }
        _log.debug('Files: %r', files)

        try:
            response = await http.post(
                url=self.get_upload_url(),
                cache=False,
                user_agent=True,
                data=post_data,
                files=files,
            )
        except errors.RequestError as e:
            _log.debug(f'Request failed: {e!r}')
            _log.debug(f'url={e.url!r}')
            _log.debug(f'text={e.text!r}')
            _log.debug(f'headers={e.headers!r}')
            _log.debug(f'status_code={e.status_code!r}')
            # The error message in the HTTP response is JSON. Try to parse that
            # to get the actual error message. If that fails, raise the
            # RequestError as is.
            json = e.json(default=False)
            if not json:
                raise errors.RequestError(f'Upload failed: {e.text}')
        else:
            _log.debug('Upload response: %r', response)
            json = response.json()

        # NOTE: It seems that none of the keys in the returned JSON are
        #       guaranteed to exist.
        _log.debug('Upload JSON response: %r', json)
        _log.debug(json.get('status'))
        if json.get('status') == 'success':
            return tracker_jobs.torrent_filepath
        elif json.get('message'):
            raise errors.RequestError(f'Upload failed: {json["message"]}')
        elif json.get('error'):
            raise errors.RequestError(f'Upload failed: {json["error"]}')
        else:
            raise RuntimeError(f'Unexpected response: {json!r}')



import errno
import re
from unittest.mock import Mock, call

import pytest

from upsies import errors
from upsies.utils import string


@pytest.mark.parametrize(
    argnames='group, exp_string, exp_error',
    argvalues=(
        ('digits', '0123456789', None),
        ('foo', None, "No such character group: 'foo'"),
    ),
)
def test_group(group, exp_string, exp_error):
    if exp_error:
        with pytest.raises(ValueError, match=rf'^{re.escape(exp_error)}$'):
            string.group(group)
    else:
        assert string.group(group) == exp_string


max_nfo_size = 100

@pytest.fixture
def read_nfo_mocks(mocker):
    mocks = Mock()
    mocker.patch('upsies.utils.string._max_nfo_size', max_nfo_size)
    mocks.attach_mock(mocker.patch('upsies.utils.fs.file_list'), 'file_list')

    def mock_getsize(nfo_filepath):
        return max_nfo_size

    mocks.attach_mock(mocker.patch('os.path.getsize', side_effect=mock_getsize), 'getsize')

    def mock_open(nfo_filepath, mode):
        return Mock(
            __enter__=Mock(
                return_value=Mock(
                    read=Mock(return_value=f'content of {nfo_filepath}'.encode('utf-8')),
                ),
            ),
            __exit__=Mock(return_value=False),
        )

    mocks.attach_mock(mocker.patch('builtins.open', side_effect=mock_open), 'open')

    def mock_decode_nfo(bytes):
        return 'decoded ' + bytes.decode('utf-8')

    mocks.attach_mock(mocker.patch('upsies.utils.string.decode_nfo', side_effect=mock_decode_nfo), 'decode_nfo')

    return mocks

def test_read_nfo__no_nfos_found(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    directory = 'path/to'

    mocks.file_list.return_value = []

    return_value = string.read_nfo(directory)
    assert return_value is None
    assert mocks.mock_calls == [
        call.file_list('path/to', extensions=('nfo', 'NFO')),
    ]

def test_read_nfo__gets_filepath(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    filepath = 'path/to/specific.nfo'

    mocks.file_list.return_value = [filepath]

    def mock_open(nfo_filepath, mode):
        if 'specific.nfo' in nfo_filepath:
            return Mock(
                __enter__=Mock(
                    return_value=Mock(
                        read=Mock(return_value=f'content of {nfo_filepath}'.encode('utf-8')),
                    ),
                ),
                __exit__=Mock(return_value=False),
            )

    mocks.open.side_effect = mock_open

    return_value = string.read_nfo(filepath)
    assert return_value == f'decoded content of {filepath}'
    assert mocks.mock_calls == [
        call.file_list(filepath, extensions=('nfo', 'NFO')),
        call.getsize(filepath),
        call.open(filepath, 'rb'),
        call.decode_nfo(b'content of ' + filepath.encode('utf8')),
    ]

def test_read_nfo__getsize_does_not_raise(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    directory = 'path/to'

    mocks.file_list.return_value = [
        'path/to/unreadable.nfo',
        'path/to/toobig.nfo',
        'path/to/good.nfo',
        'path/to/surplus.nfo',
    ]

    def mock_getsize(nfo_filepath):
        if 'unreadable' in nfo_filepath:
            raise OSError(errno.EACCES, 'Permission denied')
        elif 'toobig' in nfo_filepath:
            return max_nfo_size + 1
        else:
            return max_nfo_size

    mocks.getsize.side_effect = mock_getsize

    return_value = string.read_nfo(directory)
    assert return_value == 'decoded content of path/to/good.nfo'
    assert mocks.mock_calls == [
        call.file_list('path/to', extensions=('nfo', 'NFO')),
        call.getsize('path/to/unreadable.nfo'),
        call.getsize('path/to/toobig.nfo'),
        call.getsize('path/to/good.nfo'),
        call.open('path/to/good.nfo', 'rb'),
        call.decode_nfo(b'content of path/to/good.nfo'),
    ]

def test_read_nfo__getsize_raises(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    directory = 'path/to'

    mocks.file_list.return_value = [
        'path/to/unreadable.nfo',
        'path/to/toobig.nfo',
    ]

    def mock_getsize(nfo_filepath):
        if 'unreadable' in nfo_filepath:
            raise OSError(errno.EACCES, 'Permission denied')
        elif 'toobig' in nfo_filepath:
            return max_nfo_size + 1
        else:
            return max_nfo_size

    mocks.getsize.side_effect = mock_getsize

    with pytest.raises(errors.ContentError, match=r'^Failed to read nfo: Permission denied$'):
        string.read_nfo(directory)

    assert mocks.mock_calls == [
        call.file_list('path/to', extensions=('nfo', 'NFO')),
        call.getsize('path/to/unreadable.nfo'),
        call.getsize('path/to/toobig.nfo'),
    ]

def test_read_nfo__open_does_not_raise(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    directory = 'path/to'

    mocks.file_list.return_value = [
        'path/to/unreadable.nfo',
        'path/to/good.nfo',
        'path/to/surplus.nfo',
    ]

    def mock_open(nfo_filepath, mode):
        if 'unreadable' in nfo_filepath:
            raise OSError(errno.EACCES, 'Permission denied')
        else:
            return Mock(
                __enter__=Mock(
                    return_value=Mock(
                        read=Mock(return_value=f'content of {nfo_filepath}'.encode('utf-8')),
                    ),
                ),
                __exit__=Mock(return_value=False),
            )

    mocks.open.side_effect = mock_open

    return_value = string.read_nfo(directory)
    assert return_value == 'decoded content of path/to/good.nfo'
    assert mocks.mock_calls == [
        call.file_list('path/to', extensions=('nfo', 'NFO')),
        call.getsize('path/to/unreadable.nfo'),
        call.open('path/to/unreadable.nfo', 'rb'),
        call.getsize('path/to/good.nfo'),
        call.open('path/to/good.nfo', 'rb'),
        call.decode_nfo(b'content of path/to/good.nfo'),
    ]


def test_read_nfo__open_raises(read_nfo_mocks, mocker):
    mocks = read_nfo_mocks
    directory = 'path/to'

    mocks.file_list.return_value = [
        'path/to/unreadable.nfo',
    ]

    def mock_open(nfo_filepath, mode):
        if 'unreadable' in nfo_filepath:
            raise OSError(errno.EACCES, 'Permission denied')
        else:
            return Mock(
                __enter__=Mock(
                    return_value=Mock(
                        read=Mock(return_value=f'content of {nfo_filepath}'.encode('utf-8')),
                    ),
                ),
                __exit__=Mock(return_value=False),
            )

    mocks.open.side_effect = mock_open

    with pytest.raises(errors.ContentError, match=r'^Failed to read nfo: Permission denied$'):
        string.read_nfo(directory)

    assert mocks.mock_calls == [
        call.file_list('path/to', extensions=('nfo', 'NFO')),
        call.getsize('path/to/unreadable.nfo'),
        call.open('path/to/unreadable.nfo', 'rb'),
    ]


@pytest.mark.parametrize(
    argnames='bytes, exp_string',
    argvalues=(
        (b'M\x94\x94', 'Möö'),  # CP437
        (b'M\xC3\xB6\xC3\xB6', 'Möö'),  # UTF-8
    ),
)
def test_decode_nfo(bytes, exp_string):
    assert string.decode_nfo(bytes) == exp_string


@pytest.mark.parametrize(
    argnames='text, exp_text',
    argvalues=(
        ('THIS is a sTRING', 'This Is A String'),
        ("this is jeremy's string", "This Is Jeremy's String"),
        (' this\tstring  has   weird    spacing  ', ' This\tString  Has   Weird    Spacing  '),
    ),
)
def test_capitalize(text, exp_text):
    assert string.capitalize(text) == exp_text


@pytest.mark.parametrize(
    argnames='ratings, exp_string',
    argvalues=(
        ((-0.0, -0.1, -10.5), '☆☆☆☆☆☆☆☆☆☆'),
        ((0, 0.1, 0.2, 0.3), '☆☆☆☆☆☆☆☆☆☆'), ((0.4, 0.5, 0.6), '⯪☆☆☆☆☆☆☆☆☆'), ((0.7, 0.8, 0.9), '★☆☆☆☆☆☆☆☆☆'),
        ((1, 1.1, 1.2, 1.3), '★☆☆☆☆☆☆☆☆☆'), ((1.4, 1.5, 1.6), '★⯪☆☆☆☆☆☆☆☆'), ((1.7, 1.8, 1.9), '★★☆☆☆☆☆☆☆☆'),
        ((2, 2.1, 2.2, 2.3), '★★☆☆☆☆☆☆☆☆'), ((2.4, 2.5, 2.6), '★★⯪☆☆☆☆☆☆☆'), ((2.7, 2.8, 2.9), '★★★☆☆☆☆☆☆☆'),
        ((3, 3.1, 3.2, 3.3), '★★★☆☆☆☆☆☆☆'), ((3.4, 3.5, 3.6), '★★★⯪☆☆☆☆☆☆'), ((3.7, 3.8, 3.9), '★★★★☆☆☆☆☆☆'),
        ((4, 4.1, 4.2, 4.3), '★★★★☆☆☆☆☆☆'), ((4.4, 4.5, 4.6), '★★★★⯪☆☆☆☆☆'), ((4.7, 4.8, 4.9), '★★★★★☆☆☆☆☆'),
        ((5, 5.1, 5.2, 5.3), '★★★★★☆☆☆☆☆'), ((5.4, 5.5, 5.6), '★★★★★⯪☆☆☆☆'), ((5.7, 5.8, 5.9), '★★★★★★☆☆☆☆'),
        ((6, 6.1, 6.2, 6.3), '★★★★★★☆☆☆☆'), ((6.4, 6.5, 6.6), '★★★★★★⯪☆☆☆'), ((6.7, 6.8, 6.9), '★★★★★★★☆☆☆'),
        ((7, 7.1, 7.2, 7.3), '★★★★★★★☆☆☆'), ((7.4, 7.5, 7.6), '★★★★★★★⯪☆☆'), ((7.7, 7.8, 7.9), '★★★★★★★★☆☆'),
        ((8, 8.1, 8.2, 8.3), '★★★★★★★★☆☆'), ((8.4, 8.5, 8.6), '★★★★★★★★⯪☆'), ((8.7, 8.8, 8.9), '★★★★★★★★★☆'),
        ((9, 9.1, 9.2, 9.3), '★★★★★★★★★☆'), ((9.4, 9.5, 9.6), '★★★★★★★★★⯪'), ((9.7, 9.8, 9.9), '★★★★★★★★★★'),
        ((10, 10.1, 10.5, 12), '★★★★★★★★★★'),
    ),
)
def test_star_rating(ratings, exp_string):
    for rating in ratings:
        assert string.star_rating(rating) == exp_string


def test_remove_prefix():
    assert string.remove_prefix('com.domain.www', 'com.') == 'domain.www'
    assert string.remove_prefix('com.domain.www', 'moc.') == 'com.domain.www'


def test_remove_suffix():
    assert string.remove_suffix('www.domain.com', '.com') == 'www.domain'
    assert string.remove_suffix('www.domain.com', '.moc') == 'www.domain.com'


@pytest.mark.parametrize(
    argnames='fstring, local_objects, exp_result',
    argvalues=(
        ('foo {add(1, 2, 3) * 2} bar', {'add': lambda *n: sum(n)}, 'foo 12 bar'),
        ('foo {add(1, 2, 3) * 2} bar', {}, NameError("name 'add' is not defined")),
    ),
)
def test_evaluate_fstring(fstring, local_objects, exp_result):
    locals().update(local_objects)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            string.evaluate_fstring(fstring)
    else:
        assert string.evaluate_fstring(fstring) == exp_result



import asyncio
from unittest.mock import Mock, call

from upsies.uis.tui import widgets


async def test_ActivityIndicator_active_argument():
    ai = widgets.ActivityIndicator(active=1)
    assert ai.active is True
    ai = widgets.ActivityIndicator(active=0)
    assert ai.active is False

async def test_ActivityIndicator_active_property(mocker):
    mocker.patch('upsies.uis.tui.widgets.ActivityIndicator._iterate')
    ai = widgets.ActivityIndicator()
    assert ai.active is False
    assert ai._iterate.call_args_list == []
    ai.active = 1
    assert ai.active is True
    assert ai._iterate.call_args_list == [call()]
    ai.active = 0
    assert ai.active is False
    assert ai._iterate.call_args_list == [call()]
    ai.active = 'yes'
    assert ai.active is True
    assert ai._iterate.call_args_list == [call(), call()]

async def test_ActivityIndicator_active_property_calls_iterate(mocker):
    ai = widgets.ActivityIndicator(interval=0)
    mocker.patch.object(ai, '_iterate')
    assert ai._iterate.call_args_list == []
    ai.active = True
    assert ai._iterate.call_args_list == [call()]


async def test_ActivityIndicator_enable_disable(mocker):
    ai = widgets.ActivityIndicator(interval=0)
    mocker.patch.object(ai, '_iterate')
    assert ai.active is False
    ai.enable()
    assert ai.active is True
    ai.disable()
    assert ai.active is False


async def test_ActivityIndicator_format_property(mocker):
    mocker.patch('upsies.uis.tui.widgets.ActivityIndicator._iterate')
    ai = widgets.ActivityIndicator()
    assert ai.format == '{indicator}'
    ai.format = '[{indicator}]'
    assert ai.format == '[{indicator}]'


async def test_ActivityIndicator_text_property(mocker):
    event_loop = asyncio.get_running_loop()
    ai = widgets.ActivityIndicator(states=('a', 'b', 'c'), format=':{indicator}:', active=True)
    call_later_mock = mocker.patch.object(event_loop, 'call_later')
    assert ai.text == ':a:'
    ai._iterate()
    assert ai.text == ':b:'
    assert call_later_mock.call_args_list == [
        call(ai._interval, ai._iterate),
    ]

    ai._iterate()
    assert ai.text == ':c:'
    assert call_later_mock.call_args_list == [
        call(ai._interval, ai._iterate), call(ai._interval, ai._iterate),
    ]

    ai._iterate()
    assert ai.text == ':a:'
    assert call_later_mock.call_args_list == [
        call(ai._interval, ai._iterate), call(ai._interval, ai._iterate), call(ai._interval, ai._iterate),
    ]


async def test_ActivityIndicator_iterate_while_not_active(mocker):
    event_loop = asyncio.get_running_loop()
    cb = Mock()
    ai = widgets.ActivityIndicator(callback=cb, interval=123, states=('a', 'b', 'c'))
    call_later_mock = mocker.patch.object(event_loop, 'call_later')
    ai._iterate()
    assert cb.call_args_list == []
    assert call_later_mock.call_args_list == []
    ai._iterate()
    assert cb.call_args_list == []
    assert call_later_mock.call_args_list == []
    ai._iterate()
    assert cb.call_args_list == []
    assert call_later_mock.call_args_list == []


async def test_ActivityIndicator_iterate_while_active(mocker):
    event_loop = asyncio.get_running_loop()
    cb = Mock()
    ai = widgets.ActivityIndicator(callback=cb, interval=123, states=('a', 'b', 'c'),
                                   active=False, format='<{indicator}>')
    call_later_mock = mocker.patch.object(event_loop, 'call_later')
    assert cb.call_args_list == []
    assert call_later_mock.call_args_list == []
    ai.active = True
    assert cb.call_args_list == [
        call('<a>'),
    ]
    assert call_later_mock.call_args_list == [
        call(123.0, ai._iterate),
    ]
    ai._iterate()
    assert cb.call_args_list == [
        call('<a>'),
        call('<b>'),
    ]
    assert call_later_mock.call_args_list == [
        call(123.0, ai._iterate),
        call(123.0, ai._iterate),
    ]
    ai._iterate()
    assert cb.call_args_list == [
        call('<a>'),
        call('<b>'),
        call('<c>'),
    ]
    assert call_later_mock.call_args_list == [
        call(123.0, ai._iterate),
        call(123.0, ai._iterate),
        call(123.0, ai._iterate),
    ]



import re
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies import errors, utils
from upsies.trackers import ptp


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.name = 'ptp'
    return tracker


@pytest.fixture
def imghost():
    class MockImageHost(utils.imghosts.base.ImageHostBase):
        name = 'mock image host'
        default_config = {}
        _upload_image = AsyncMock()

    return MockImageHost()


@pytest.fixture
def ptp_tracker_jobs(imghost, tracker, tmp_path, mocker):
    content_path = tmp_path / 'Foo 2000 1080p BluRay x264-ASDF'

    ptp_tracker_jobs = ptp.PtpTrackerJobs(
        content_path=str(content_path),
        tracker=tracker,
        image_hosts=(imghost,),
        btclient=Mock(),
        torrent_destination=str(tmp_path / 'destination'),
        common_job_args={
            'home_directory': str(tmp_path / 'home_directory'),
            'ignore_cache': True,
        },
        options=None,
    )

    return ptp_tracker_jobs


@pytest.fixture
def mock_job_attributes(mocker):
    def mock_job_attributes(ptp_tracker_jobs):
        job_attrs = (
            # Common background jobs
            'create_torrent_job',
            'mediainfo_job',
            'screenshots_job',
            'upload_screenshots_job',
            'ptp_group_id_job',
            'description_job',
            'trumpable_job',

            # Common interactive jobs
            'type_job',
            'imdb_job',
            'source_job',
            'scene_check_job',

            # Interactive jobs that only run if movie does not exists on PTP yet
            'title_job',
            'year_job',
            'plot_job',
            'tags_job',
            'cover_art_job',
        )
        for job_attr in job_attrs:
            mocker.patch.object(
                type(ptp_tracker_jobs),
                job_attr,
                PropertyMock(return_value=Mock(attr=job_attr)),
            )

    return mock_job_attributes


def test_jobs_before_upload_items(ptp_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(ptp_tracker_jobs)

    assert tuple(job.attr for job in ptp_tracker_jobs.jobs_before_upload) == (
        # Common background jobs
        'create_torrent_job',
        'mediainfo_job',
        'screenshots_job',
        'upload_screenshots_job',
        'ptp_group_id_job',
        'description_job',
        'trumpable_job',

        # Common interactive jobs
        'type_job',
        'imdb_job',
        'source_job',
        'scene_check_job',

        # Interactive jobs that only run if movie does not exists on PTP yet
        'title_job',
        'year_job',
        'plot_job',
        'tags_job',
        'cover_art_job',
    )


def test_isolated_jobs__only_description(ptp_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(ptp_tracker_jobs)
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value={'only_description': True}))
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    assert ptp_tracker_jobs.isolated_jobs is ptp_tracker_jobs.get_job_and_dependencies.return_value
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(ptp_tracker_jobs.description_job, ptp_tracker_jobs.screenshots_job)
    ]

def test_isolated_jobs__no_isolated_jobs(ptp_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(ptp_tracker_jobs)
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value={}))
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    assert ptp_tracker_jobs.isolated_jobs == ()
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == []


def test_type_job(ptp_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.type_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Type',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        autodetect=ptp_tracker_jobs.autodetect_type,
        options=(
            ('Feature Film', 'Feature Film'),
            ('Short Film', 'Short Film'),
            ('Miniseries', 'Miniseries'),
            ('Stand-up Comedy', 'Stand-up Comedy'),
            ('Live Performance', 'Live Performance'),
            ('Movie Collection', 'Movie Collection'),
        ),
        callbacks={
            'finished': ptp_tracker_jobs.update_imdb_query,
        },
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('type')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call('type_job')]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='release_type, main_video_duration, exp_return_value, exp_duration_calls',
    argvalues=(
        (utils.release.ReleaseType.season, 123, 'Miniseries', False),
        (utils.release.ReleaseType.movie, 45 * 60, 'Short Film', True),
        (utils.release.ReleaseType.movie, (45 * 60) + 1, None, True),
    ),
    ids=lambda v: str(v),
)
def test_autodetect_type(
        release_type, main_video_duration,
        exp_return_value, exp_duration_calls,
        ptp_tracker_jobs, mocker,
):
    mocker.patch.object(type(ptp_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        type=release_type,
    )))

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.find_videos', return_value=iter(
            ('Foo.S01E01.mkv', 'Foo.S01E02.mkv', 'Foo.S01E03.mkv', 'Foo.sample.mkv'),
        )),
        'find_videos',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.filter_main_videos', return_value=iter(
            ('Foo.S01E01.mkv', 'Foo.S01E02.mkv', 'Foo.S01E03.mkv'),
        )),
        'filter_main_videos',
    )
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.duration', return_value=main_video_duration),
        'duration',
    )

    return_value = ptp_tracker_jobs.autodetect_type('job_')
    assert return_value == exp_return_value

    if exp_duration_calls:
        assert mocks.mock_calls == [
            call.find_videos(ptp_tracker_jobs.content_path),
            call.filter_main_videos(('Foo.S01E01.mkv', 'Foo.S01E02.mkv', 'Foo.S01E03.mkv', 'Foo.sample.mkv')),
            call.duration('Foo.S01E01.mkv'),
        ]
    else:
        assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames='type_job_output, exp_query_type',
    argvalues=(
        ([], '<original query type>'),
        (['Feature Film'], utils.release.ReleaseType.movie),
        (['Miniseries'], utils.release.ReleaseType.season),
        (['<anything else>'], '<original query type>'),
    ),
    ids=lambda v: str(v),
)
def test_update_imdb_query(type_job_output, exp_query_type, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(
        query=Mock(
            type='<original query type>',
        ),
    )))
    mocker.patch.object(type(ptp_tracker_jobs), 'type_job', PropertyMock(return_value=Mock(
        is_finished=True,
        output=type_job_output,
    )))

    return_value = ptp_tracker_jobs.update_imdb_query('job_')
    assert return_value is None
    assert ptp_tracker_jobs.imdb_job.query.type == exp_query_type


def test_trumpable_job(ptp_tracker_jobs, mocker):
    CustomJob_mock = mocker.patch('upsies.jobs.custom.CustomJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.trumpable_job is CustomJob_mock.return_value
    assert CustomJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Trumpable',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        worker=ptp_tracker_jobs.autodetect_trumpable,
        no_output_is_ok=True,
        catch=(
            errors.ContentError,
        ),
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('trumpable')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call('trumpable_job')]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.parametrize(
    argnames='options, exp_hardcoded_subtitles_from_options, exp_no_english_subtitles_from_options',
    argvalues=(
        ({}, None, None),
        ({'hardcoded_subtitles': None, 'no_english_subtitles': None}, None, None),
        ({'hardcoded_subtitles': None, 'no_english_subtitles': True}, None, True),
        ({'hardcoded_subtitles': None, 'no_english_subtitles': False}, None, False),
        ({'hardcoded_subtitles': True, 'no_english_subtitles': None}, True, None),
        ({'hardcoded_subtitles': False, 'no_english_subtitles': None}, False, None),
        ({'hardcoded_subtitles': False, 'no_english_subtitles': True}, False, True),
        ({'hardcoded_subtitles': True, 'no_english_subtitles': False}, True, False),
        ({'hardcoded_subtitles': True, 'no_english_subtitles': True}, True, True),
        ({'hardcoded_subtitles': False, 'no_english_subtitles': False}, False, False),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.parametrize(
    argnames='audio_languages, subtitle_languages, exp_no_english_subtitles_from_autodetection',
    argvalues=(
        pytest.param(
            None, None, False,
            id='Track detection fails',
        ),
        pytest.param(
            set(), set(), False,
            id='No audio tracks + No subtitle tracks',
        ),
        pytest.param(
            {'se'},
            {},
            True,
            id='No English audio + No subtitles',
        ),
        pytest.param(
            {'se'},
            {'se', 'en', 'dk', 'no'},
            False,
            id='No English audio + English subtitles',
        ),
        pytest.param(
            {'se', 'en'},
            {'se', 'dk', 'no'},
            False,
            id='English audio + No English subtitles',
        ),
        pytest.param(
            {'se', 'en'},
            {'en', 'se', 'dk', 'no'},
            False,
            id='English audio + English subtitles',
        ),
        pytest.param(
            {'se', ''},
            {'se', 'dk', 'no'},
            False,
            id='No English audio + No English subtitles + Unknown audio',
        ),
        pytest.param(
            {'se'},
            {'se', '', 'dk', 'no'},
            False,
            id='No English audio + No English subtitles + Unknown subtitles',
        ),
    ),
    ids=lambda v: str(v),
)
async def test_autodetect_trumpable(
        options,
        exp_hardcoded_subtitles_from_options, exp_no_english_subtitles_from_options,
        audio_languages, subtitle_languages,
        exp_no_english_subtitles_from_autodetection,
        ptp_tracker_jobs, mocker,
):
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value=options))

    languages_mock = mocker.patch('upsies.utils.video.languages', return_value=(
        {'Audio': audio_languages}
        if audio_languages is not None else
        {}
    ))

    mocker.patch.object(type(ptp_tracker_jobs), 'subtitles', PropertyMock(return_value=(
        [
            Mock(language=language)
            for language in subtitle_languages
        ]
        if subtitle_languages is not None else
        []
    )))

    reasons = await ptp_tracker_jobs.autodetect_trumpable('job_')

    exp_hardcoded_subtitles = (
        exp_hardcoded_subtitles_from_options
        if exp_hardcoded_subtitles_from_options is not None else
        False
    )

    exp_no_english_subtitles = (
        exp_no_english_subtitles_from_options
        if exp_no_english_subtitles_from_options is not None else
        exp_no_english_subtitles_from_autodetection
    )

    exp_reasons = []
    if exp_hardcoded_subtitles:
        exp_reasons.append(ptp.metadata.PtpTrumpableReason.HARDCODED_SUBTITLES)
    if exp_no_english_subtitles:
        exp_reasons.append(ptp.metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES)

    assert reasons == exp_reasons

    if exp_no_english_subtitles_from_options is None:
        assert languages_mock.call_args_list == [
            call(ptp_tracker_jobs.content_path, default=''),
        ]


def test_description_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'mediainfo_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'upload_screenshots_job', PropertyMock())
    mocker.patch.object(ptp_tracker_jobs, 'generate_description')
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.description_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Description',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.generate_description,
        read_only=True,
        hidden=True,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('description')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call('description_job')]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.mediainfo_job,
        ptp_tracker_jobs.upload_screenshots_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


@pytest.mark.parametrize(
    argnames='content_path, exp_original_release_name',
    argvalues=(
        ('path/to/Foo.2002.720p.x264-ASDF.mkv', 'Foo.2002.720p.x264-ASDF'),
        ('path/to/Foo.2002.720p.x264-ASDF', 'Foo.2002.720p.x264-ASDF'),
        ('Foo.2002.720p.x264-ASDF.mkv', 'Foo.2002.720p.x264-ASDF'),
        ('Foo.2002.720p.x264-ASDF', 'Foo.2002.720p.x264-ASDF'),
    ),
    ids=lambda v: str(v),
)
async def test_generate_description(content_path, exp_original_release_name, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'content_path', PropertyMock(return_value=content_path))
    mocker.patch.object(type(ptp_tracker_jobs), 'mediainfos_and_screenshots', PropertyMock(return_value={
        'path/to/video1.mkv': {
            'mediainfo': '<mediainfo of path/to/video1.mkv>',
            'screenshot_urls': (
                'http://screenshot1.of.video1.mkv',
                'http://screenshot2.of.video1.mkv',
                'http://screenshot3.of.video1.mkv',
            ),
        },
        'path/to/video2.mkv': {
            'mediainfo': None,
            'screenshot_urls': (
                'http://screenshot1.of.video2.mkv',
                'http://screenshot2.of.video2.mkv',
            ),
        },
        'path/to/video3.mkv': {
            'mediainfo': '<mediainfo of path/to/video3.mkv>',
            'screenshot_urls': (),
        },
        'path/to/video4.mkv': {
            'mediainfo': None,
            'screenshot_urls': (),
        },
    }))

    description = await ptp_tracker_jobs.generate_description()
    assert description == (
        f'[size=4][b]{exp_original_release_name}[/b][/size]'
        '\n\n'
        '[mediainfo]<mediainfo of path/to/video1.mkv>[/mediainfo]'
        '[img=http://screenshot1.of.video1.mkv]\n'
        '[img=http://screenshot2.of.video1.mkv]\n'
        '[img=http://screenshot3.of.video1.mkv]'
        '\n[hr]\n'
        '[img=http://screenshot1.of.video2.mkv]\n'
        '[img=http://screenshot2.of.video2.mkv]'
    )


def test_mediainfo_from_all_videos(ptp_tracker_jobs):
    assert ptp_tracker_jobs.mediainfo_from_all_videos is True


def test_screenshots_from_all_videos(ptp_tracker_jobs):
    assert ptp_tracker_jobs.screenshots_from_all_videos is True


@pytest.mark.parametrize(
    argnames='options, filtered_video_files, exp_screenshots_count, exp_filter_calls',
    argvalues=(
        pytest.param(
            {'screenshots': 123, 'screenshots_from_movie': 456, 'screenshots_from_episode': 789},
            None,
            123,
            False,
            id='User-provided number of screenshots',
        ),
        pytest.param(
            {'screenshots_from_movie': 456, 'screenshots_from_episode': 789},
            ['file1.mkv'],
            456,
            True,
            id='Number of screenshots for movies',
        ),
        pytest.param(
            {'screenshots_from_movie': 456, 'screenshots_from_episode': 789},
            ['foo1.mkv', 'foo2.mkv'],
            789,
            True,
            id='Number of screenshots for episodes',
        ),
    ),
    ids=lambda v: str(v),
)
def test_screenshots_count(
        options, filtered_video_files,
        exp_screenshots_count, exp_filter_calls,
        ptp_tracker_jobs, mocker,
):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.utils.torrent.filter_files'), 'filter_files')
    mocks.attach_mock(
        mocker.patch('upsies.utils.video.filter_main_videos', return_value=filtered_video_files),
        'filter_main_videos',
    )
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value=options))
    mocker.patch.object(type(ptp_tracker_jobs), 'content_path', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'exclude_files', PropertyMock())

    screenshots_count = ptp_tracker_jobs.screenshots_count
    assert screenshots_count == exp_screenshots_count

    if exp_filter_calls:
        assert mocks.mock_calls == [
            call.filter_files(ptp_tracker_jobs.content_path, exclude=ptp_tracker_jobs.exclude_files),
            call.filter_main_videos(mocks.filter_files.return_value),
        ]
    else:
        assert mocks.mock_calls == []


@pytest.mark.parametrize(
    argnames='imdb_job, exp_imdb_id',
    argvalues=(
        (Mock(is_finished=False), None),
        (Mock(is_finished=True, output=['tt123456']), 'tt123456'),
    ),
    ids=lambda v: str(v),
)
def test_imdb_id(imdb_job, exp_imdb_id, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    imdb_id = ptp_tracker_jobs.imdb_id
    assert imdb_id == exp_imdb_id


def test_ptp_group_id_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    mocker.patch.object(ptp_tracker_jobs, 'fetch_ptp_group_id')
    CustomJob_mock = mocker.patch('upsies.jobs.custom.CustomJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.ptp_group_id_job is CustomJob_mock.return_value
    assert CustomJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='PTP Group ID',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        worker=ptp_tracker_jobs.fetch_ptp_group_id,
        catch=(
            errors.RequestError,
        ),
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('ptp-group-id')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call('ptp_group_id_job')]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.imdb_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='group_id, exp_group_id',
    argvalues=(
        (None, ''),
        ('123456', '123456'),
    ),
    ids=lambda v: str(v),
)
async def test_fetch_ptp_group_id(group_id, exp_group_id, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(is_finished=True)))
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_id', PropertyMock(return_value='tt123456'))
    mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_group_id_by_imdb_id', AsyncMock(return_value=group_id))

    group_id = await ptp_tracker_jobs.fetch_ptp_group_id('ignored job')
    assert group_id == exp_group_id

    assert ptp_tracker_jobs.tracker.get_ptp_group_id_by_imdb_id.call_args_list == [call(ptp_tracker_jobs.imdb_id)]


@pytest.mark.parametrize(
    argnames='ptp_group_id_job, exp_ptp_group_id',
    argvalues=(
        (Mock(is_finished=False), None),
        (Mock(is_finished=True, output=[]), None),
        (Mock(is_finished=True, output=['123']), '123'),
    ),
    ids=lambda v: str(v),
)
async def test_ptp_group_id(ptp_group_id_job, exp_ptp_group_id, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock(return_value=ptp_group_id_job))
    ptp_group_id = ptp_tracker_jobs.ptp_group_id
    assert ptp_group_id == exp_ptp_group_id


@pytest.mark.parametrize(
    argnames='ptp_group_id_job, exp_ptp_group_does_not_exist',
    argvalues=(
        (Mock(is_finished=False), AssertionError),
        (Mock(is_finished=True, output=[]), True),
        (Mock(is_finished=True, output=['123']), False),
    ),
    ids=lambda v: str(v),
)
async def test_ptp_group_does_not_exist(ptp_group_id_job, exp_ptp_group_does_not_exist, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock(return_value=ptp_group_id_job))
    if (
            type(exp_ptp_group_does_not_exist) is type
            and issubclass(exp_ptp_group_does_not_exist, Exception)
    ):
        with pytest.raises(exp_ptp_group_does_not_exist) as exception_info:
            ptp_tracker_jobs.ptp_group_does_not_exist()
        assert exception_info.value.args[0] is ptp_tracker_jobs.ptp_group_id_job
    else:
        ptp_group_does_not_exist = ptp_tracker_jobs.ptp_group_does_not_exist()
        assert ptp_group_does_not_exist == exp_ptp_group_does_not_exist


def test_title_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_does_not_exist')
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.title_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Title',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.fetch_title,
        normalizer=ptp_tracker_jobs.normalize_title,
        validator=ptp_tracker_jobs.validate_title,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('title')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call(
        'title_job',
        precondition=ptp_tracker_jobs.ptp_group_does_not_exist,
    )]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.ptp_group_id_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='imdb_job, exp_exception',
    argvalues=(
        (Mock(is_finished=False), AssertionError),
        (Mock(is_finished=True, output=['123']), None),
    ),
    ids=lambda v: str(v),
)
async def test_fetch_title(imdb_job, exp_exception, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_metadata', AsyncMock())

    if exp_exception:
        with pytest.raises(exp_exception) as exception_info:
            await ptp_tracker_jobs.fetch_title()
        assert exception_info.value.args[0] is ptp_tracker_jobs.imdb_job
    else:
        return_value = await ptp_tracker_jobs.fetch_title()
        assert return_value == ptp_tracker_jobs.tracker.get_ptp_metadata.return_value
        ptp_tracker_jobs.tracker.get_ptp_metadata.call_args_list == [
            call(ptp_tracker_jobs.imdb_id, key='title'),
        ]


@pytest.mark.parametrize(
    argnames='text, exp_title',
    argvalues=(
        ('The Title', 'The Title'),
        (' The Title', 'The Title'),
        (' The Title  ', 'The Title'),
        ('\nThe Title\t', 'The Title'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_title(text, exp_title, ptp_tracker_jobs, mocker):
    return_value = ptp_tracker_jobs.normalize_title(text)
    assert return_value == exp_title


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', ValueError('Title must not be empty.')),
        ('not empty', None),
    ),
    ids=lambda v: str(v),
)
async def test_validate_title(text, exp_exception, ptp_tracker_jobs, mocker):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            ptp_tracker_jobs.validate_title(text)
    else:
        return_value = ptp_tracker_jobs.validate_title(text)
        assert return_value is None


def test_year_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_does_not_exist')
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.year_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Year',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.fetch_year,
        normalizer=ptp_tracker_jobs.normalize_year,
        validator=ptp_tracker_jobs.validate_year,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('year')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call(
        'year_job',
        precondition=ptp_tracker_jobs.ptp_group_does_not_exist,
    )]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.ptp_group_id_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='imdb_job, exp_exception',
    argvalues=(
        (Mock(is_finished=False), AssertionError),
        (Mock(is_finished=True, output=['2012']), None),
    ),
    ids=lambda v: str(v),
)
async def test_fetch_year(imdb_job, exp_exception, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_metadata', AsyncMock())

    if exp_exception:
        with pytest.raises(exp_exception) as exception_info:
            await ptp_tracker_jobs.fetch_year()
        assert exception_info.value.args[0] is ptp_tracker_jobs.imdb_job
    else:
        return_value = await ptp_tracker_jobs.fetch_year()
        assert return_value == ptp_tracker_jobs.tracker.get_ptp_metadata.return_value
        ptp_tracker_jobs.tracker.get_ptp_metadata.call_args_list == [
            call(ptp_tracker_jobs.imdb_id, key='year'),
        ]


@pytest.mark.parametrize(
    argnames='text, exp_year',
    argvalues=(
        ('2012', '2012'),
        (' 2012', '2012'),
        (' 2012  ', '2012'),
        ('\n2012\t', '2012'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_year(text, exp_year, ptp_tracker_jobs, mocker):
    return_value = ptp_tracker_jobs.normalize_year(text)
    assert return_value == exp_year


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', ValueError('Year must not be empty.')),
        ('not empty', ValueError('Year is not a number.')),
        ('2012.3', ValueError('Year is not a number.')),
        ('1200', ValueError('Year is not reasonable.')),
        ('2012', None),
    ),
    ids=lambda v: str(v),
)
async def test_validate_year(text, exp_exception, ptp_tracker_jobs, mocker):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            ptp_tracker_jobs.validate_year(text)
    else:
        return_value = ptp_tracker_jobs.validate_year(text)
        assert return_value is None


def test_tags_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_does_not_exist')
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.tags_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Tags',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.fetch_tags,
        normalizer=ptp_tracker_jobs.normalize_tags,
        validator=ptp_tracker_jobs.validate_tags,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('tags')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call(
        'tags_job',
        precondition=ptp_tracker_jobs.ptp_group_does_not_exist,
    )]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.ptp_group_id_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='imdb_job, exp_exception',
    argvalues=(
        (Mock(is_finished=False), AssertionError),
        (Mock(is_finished=True, output=['foo,bar,baz']), None),
    ),
    ids=lambda v: str(v),
)
async def test_fetch_tags(imdb_job, exp_exception, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_metadata', AsyncMock())

    if exp_exception:
        with pytest.raises(exp_exception) as exception_info:
            await ptp_tracker_jobs.fetch_tags()
        assert exception_info.value.args[0] is ptp_tracker_jobs.imdb_job
    else:
        return_value = await ptp_tracker_jobs.fetch_tags()
        assert return_value == ptp_tracker_jobs.tracker.get_ptp_metadata.return_value
        ptp_tracker_jobs.tracker.get_ptp_metadata.call_args_list == [
            call(ptp_tracker_jobs.imdb_id, key='tags'),
        ]


@pytest.mark.parametrize(
    argnames='text, exp_tags',
    argvalues=(
        pytest.param('foo,bar,baz', 'bar, baz, foo', id='Tags are sorted'),
        pytest.param('foo,bar,baz,foo', 'bar, baz, foo', id='Tags are deduplicated'),
        pytest.param(',,,foo,,bar,baz,', 'bar, baz, foo', id='Empty tags are ignored'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_tags(text, exp_tags, ptp_tracker_jobs, mocker):
    return_value = ptp_tracker_jobs.normalize_tags(text)
    assert return_value == exp_tags

    def normalize_tags(self, text):
        tags = [tag.strip() for tag in text.split(',')]
        deduped = list(dict.fromkeys(tags))
        return ', '.join(tag for tag in deduped if tag)


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('horror,awesome,drama', ValueError('Tag is not valid: awesome')),
        ('horror,drama', None),
        ('', None),
    ),
    ids=lambda v: str(v),
)
async def test_validate_tags(text, exp_exception, ptp_tracker_jobs, mocker):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            ptp_tracker_jobs.validate_tags(text)
    else:
        return_value = ptp_tracker_jobs.validate_tags(text)
        assert return_value is None


def test_cover_art_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_does_not_exist')
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.cover_art_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Cover Art',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.get_cover_art,
        validator=ptp_tracker_jobs.validate_cover_art,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('cover-art')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call(
        'cover_art_job',
        precondition=ptp_tracker_jobs.ptp_group_does_not_exist,
    )]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.ptp_group_id_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='text, exp_result',
    argvalues=(
        ('http://foo.jpg', None),
        ('https://foo.jpg', None),
        ('HTTPS://foo.jpg', None),
        ('foo.jpg', ValueError('Must be an http(s):// URL.')),
    ),
    ids=lambda v: str(v),
)
def test_validate_cover_art(text, exp_result, ptp_tracker_jobs):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            ptp_tracker_jobs.validate_cover_art(text)
    else:
        result = ptp_tracker_jobs.validate_cover_art(text)
        assert result is exp_result


@pytest.mark.parametrize(
    argnames='options, exp_method_name',
    argvalues=(
        ({}, 'get_cover_art_from_ptp'),
        ({'cover_art': None}, 'get_cover_art_from_ptp'),
        ({'cover_art': ''}, 'get_cover_art_from_ptp'),
        ({'cover_art': 'user_provided.jpg'}, 'get_cover_art_from_user'),
    ),
    ids=lambda v: str(v),
)
async def test_get_cover_art(options, exp_method_name, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value=options))
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs, 'get_cover_art_from_user'),
        'get_cover_art_from_user',
    )
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs, 'get_cover_art_from_ptp'),
        'get_cover_art_from_ptp',
    )

    result = await ptp_tracker_jobs.get_cover_art()
    exp_method = getattr(mocks, exp_method_name)
    assert result is exp_method.return_value
    assert mocks.mock_calls == [getattr(call, exp_method_name)()]


@pytest.mark.parametrize(
    argnames='options, upload_exception, exp_return_value, exp_mock_calls',
    argvalues=(
        ({}, None, '', []),
        ({'cover_art': ''}, None, '', []),
        ({'cover_art': 'http://foo.jpg'}, None, 'http://foo.jpg', []),
        ({'cover_art': 'HTTPS://foo.jpg'}, None, 'HTTPS://foo.jpg', []),
        ({'cover_art': 'foo.jpg'}, None, 'http://foo.jpg', [call.upload('foo.jpg')]),
        (
            {'cover_art': 'foo.jpg'},
            errors.RequestError('Service down'),
            '',
            [
                call.upload('foo.jpg'),
                call.error(errors.RequestError('Service down')),
            ],
        ),
    ),
    ids=lambda v: str(v),
)
async def test_get_cover_art_from_user(options, upload_exception, exp_return_value, exp_mock_calls, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value=options))
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs.image_hosts[0], 'upload', side_effect=(
            upload_exception
            if upload_exception else
            lambda f: f'http://{f}'
        )),
        'upload',
    )
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs.cover_art_job, 'error'),
        'error',
    )

    return_value = await ptp_tracker_jobs.get_cover_art_from_user()
    assert return_value == exp_return_value
    assert mocks.mock_calls == exp_mock_calls


async def test_get_cover_art_from_ptp(ptp_tracker_jobs, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(wait=AsyncMock()))),
        'wait',
    )
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_metadata', AsyncMock(return_value=Mock())),
        'get_ptp_metadata',
    )
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_id', PropertyMock(return_value='tt0123'))

    return_value = await ptp_tracker_jobs.get_cover_art_from_ptp()
    assert return_value == ptp_tracker_jobs.tracker.get_ptp_metadata.return_value
    assert mocks.mock_calls == [
        call.wait(),
        call.get_ptp_metadata('tt0123', key='art'),
    ]


def test_plot_job(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id_job', PropertyMock())
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_does_not_exist')
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock())
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_and_dependencies')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.plot_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Plot',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        prejobs=ptp_tracker_jobs.get_job_and_dependencies.return_value,
        text=ptp_tracker_jobs.fetch_plot,
        normalizer=ptp_tracker_jobs.normalize_plot,
        validator=ptp_tracker_jobs.validate_plot,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('plot')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call(
        'plot_job',
        precondition=ptp_tracker_jobs.ptp_group_does_not_exist,
    )]
    assert ptp_tracker_jobs.get_job_and_dependencies.call_args_list == [call(
        ptp_tracker_jobs.ptp_group_id_job,
    )]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='imdb_job, exp_exception',
    argvalues=(
        (Mock(is_finished=False), AssertionError),
        (Mock(is_finished=True, output=['123']), None),
    ),
    ids=lambda v: str(v),
)
async def test_fetch_plot(imdb_job, exp_exception, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_job', PropertyMock(return_value=imdb_job))
    mocker.patch.object(ptp_tracker_jobs.tracker, 'get_ptp_metadata', AsyncMock())

    if exp_exception:
        with pytest.raises(exp_exception) as exception_info:
            await ptp_tracker_jobs.fetch_plot()
        assert exception_info.value.args[0] is ptp_tracker_jobs.imdb_job
    else:
        return_value = await ptp_tracker_jobs.fetch_plot()
        assert return_value == ptp_tracker_jobs.tracker.get_ptp_metadata.return_value
        ptp_tracker_jobs.tracker.get_ptp_metadata.call_args_list == [
            call(ptp_tracker_jobs.imdb_id, key='plot'),
        ]


@pytest.mark.parametrize(
    argnames='text, exp_plot',
    argvalues=(
        ('The plot.', 'The plot.'),
        (' The plot.', 'The plot.'),
        (' The plot.  ', 'The plot.'),
        ('\nThe plot.\t', 'The plot.'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_plot(text, exp_plot, ptp_tracker_jobs, mocker):
    return_value = ptp_tracker_jobs.normalize_plot(text)
    assert return_value == exp_plot


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', ValueError('Plot must not be empty.')),
        ('not empty', None),
    ),
    ids=lambda v: str(v),
)
async def test_validate_plot(text, exp_exception, ptp_tracker_jobs, mocker):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            ptp_tracker_jobs.validate_plot(text)
    else:
        return_value = ptp_tracker_jobs.validate_plot(text)
        assert return_value is None


def test_source_job(ptp_tracker_jobs, mocker):
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(ptp_tracker_jobs, 'get_job_name')
    mocker.patch.object(ptp_tracker_jobs, 'make_precondition')
    mocker.patch.object(ptp_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert ptp_tracker_jobs.source_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=ptp_tracker_jobs.get_job_name.return_value,
        label='Source',
        precondition=ptp_tracker_jobs.make_precondition.return_value,
        text=ptp_tracker_jobs.autodetect_source,
        normalizer=ptp_tracker_jobs.normalize_source,
        validator=ptp_tracker_jobs.validate_source,
        finish_on_success=True,
        common_job_arg='common job argument',
    )]
    assert ptp_tracker_jobs.get_job_name.call_args_list == [call('source')]
    assert ptp_tracker_jobs.make_precondition.call_args_list == [call('source_job')]
    assert ptp_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='release_name_source, exp_source, exp_source_job_text',
    argvalues=(
        ('BluRay', 'Blu-ray', None),
        ('DVD', 'DVD', None),
        ('DVDRip', 'DVD', None),
        ('WEB', 'WEB', None),
        ('WEB-DL', 'WEB', None),
        ('WEBDL', 'WEB', None),
        ('WEBRip', 'WEB', None),
        ('HD-DVD', 'HD-DVD', None),
        ('HDDVD', 'HD-DVD', None),
        ('HD-TV', 'HDTV', None),
        ('HDTV', 'HDTV', None),
        ('TV', 'TV', None),
        ('VHS', 'VHS', None),
        ('VHSRip', 'VHS', None),
        ('YerMom', None, 'YerMom'),
    ),
    ids=lambda v: str(v),
)
def test_autodetect_source(release_name_source, exp_source, exp_source_job_text, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'release_name', PropertyMock(return_value=Mock(
        source=release_name_source,
    )))
    mocker.patch.object(type(ptp_tracker_jobs), 'source_job', PropertyMock(return_value=Mock(
        text='not set',
    )))

    return_value = ptp_tracker_jobs.autodetect_source()
    assert return_value == exp_source
    if exp_source_job_text is not None:
        assert ptp_tracker_jobs.source_job.text == exp_source_job_text
    else:
        assert ptp_tracker_jobs.source_job.text == 'not set'

    def autodetect_source(self):
        # Find autodetected source in valid PTP sources
        source = self.release_name.source
        for known_source, regex in self._known_sources.items():
            if regex.match(source):
                # Finish the job without prompting the user
                return known_source

        # Let the user fix the autodetected source
        self.source_job.text = source


@pytest.mark.parametrize(
    argnames='text, exp_source',
    argvalues=(
        ('blu-ray', 'Blu-ray'),
        ('Blu-ray', 'Blu-ray'),
        ('BLU-RAY', 'Blu-ray'),
        ('YerMom', 'YerMom'),
    ),
    ids=lambda v: str(v),
)
def test_normalize_source(text, exp_source, ptp_tracker_jobs, mocker):
    return_value = ptp_tracker_jobs.normalize_source(text)
    assert return_value == exp_source


@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('', ValueError('You must provide a source.')),
        ('UNKNOWN_SOURCE', ValueError('You must provide a source.')),
        ('KNOWN_SOURCE', ValueError('Source is not valid: KNOWN_SOURCE')),
        ('Blu-ray', None),
    ),
    ids=lambda v: str(v),
)
async def test_validate_source(text, exp_exception, ptp_tracker_jobs, mocker):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            ptp_tracker_jobs.validate_source(text)
    else:
        return_value = ptp_tracker_jobs.validate_source(text)
        assert return_value is None


@pytest.mark.parametrize(
    argnames='ptp_group_id, post_data_common, post_data_add_movie_format, post_data_add_new_movie, exp_post_data',
    argvalues=(
        (
            None,
            {'common': 'data', 'more': 'common values'},
            {'add': 'format', 'common': 'data for adding movie format'},
            {'new': 'movie', 'common': 'data for adding new movie'},
            {'new': 'movie', 'common': 'data for adding new movie', 'more': 'common values'},
        ),
        (
            '123456',
            {'common': 'data', 'more': 'common values'},
            {'add': 'format', 'common': 'data for adding movie format'},
            {'new': 'movie', 'common': 'data for adding new movie'},
            {'add': 'format', 'common': 'data for adding movie format', 'more': 'common values'},
        ),
    ),
    ids=lambda v: str(v),
)
async def test_post_data(
        ptp_group_id,
        post_data_common, post_data_add_movie_format, post_data_add_new_movie,
        exp_post_data,
        ptp_tracker_jobs, mocker,
):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id', PropertyMock(return_value=ptp_group_id))
    mocker.patch.object(type(ptp_tracker_jobs), '_post_data_common', PropertyMock(return_value=post_data_common))
    mocker.patch.object(type(ptp_tracker_jobs), '_post_data_add_movie_format', PropertyMock(return_value=post_data_add_movie_format))
    mocker.patch.object(type(ptp_tracker_jobs), '_post_data_add_new_movie', PropertyMock(return_value=post_data_add_new_movie))
    assert ptp_tracker_jobs.post_data == exp_post_data


@pytest.mark.parametrize('not_main_movie, exp_special', ((True, '1'), (False, None)))
@pytest.mark.parametrize('personal_rip, exp_internalrip', ((True, '1'), (False, None)))
@pytest.mark.parametrize('is_scene_release, exp_scene', ((True, '1'), (False, None)))
def test__post_data_common(
        not_main_movie, exp_special,
        personal_rip, exp_internalrip,
        is_scene_release, exp_scene,
        ptp_tracker_jobs, mock_job_attributes, mocker,
):
    mock_job_attributes(ptp_tracker_jobs)

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs, 'get_job_attribute', side_effect=(
            '<type>',
            is_scene_release,
        )),
        'get_job_attribute',
    )
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs, 'get_job_output', side_effect=(
            '<description>',
        )),
        'get_job_output',
    )
    mocker.patch.object(ptp_tracker_jobs, 'read_nfo', Mock(return_value='<nfo text>'))
    mocker.patch.object(type(ptp_tracker_jobs), 'options', PropertyMock(return_value={
        'not_main_movie': not_main_movie,
        'personal_rip': personal_rip,
        'upload_token': '<upload token>',
    }))
    for name in ('source', 'codec', 'container', 'resolution', 'remaster', 'subtitles', 'trumpable'):
        mocker.patch.object(
            type(ptp_tracker_jobs),
            f'_post_data_common_{name}',
            PropertyMock(return_value={name: 'post data'}),
        )

    exp_post_data = {
        'type': '<type>',
        'release_desc': '<description>',
        'special': exp_special,
        'internalrip': exp_internalrip,
        'scene': exp_scene,
        'nfo_text': '<nfo text>',
        'uploadtoken': '<upload token>',
        'source': 'post data',
        'codec': 'post data',
        'container': 'post data',
        'resolution': 'post data',
        'remaster': 'post data',
        'subtitles': 'post data',
        'trumpable': 'post data',
    }

    assert ptp_tracker_jobs._post_data_common == exp_post_data

    assert mocks.mock_calls == [
        call.get_job_attribute(ptp_tracker_jobs.type_job, 'choice'),
        call.get_job_output(ptp_tracker_jobs.description_job, slice=0),
        call.get_job_attribute(ptp_tracker_jobs.scene_check_job, 'is_scene_release'),
    ]


def test__post_data_common_source(ptp_tracker_jobs, mocker):
    mocker.patch.object(ptp_tracker_jobs, 'get_job_output')
    assert ptp_tracker_jobs._post_data_common_source == {
        'source': 'Other',
        'other_source': ptp_tracker_jobs.get_job_output.return_value,
    }
    assert ptp_tracker_jobs.get_job_output.call_args_list == [
        call(ptp_tracker_jobs.source_job, slice=0),
    ]


def test__post_data_common_codec(ptp_tracker_jobs, mocker):
    assert ptp_tracker_jobs._post_data_common_codec == {
        'codec': '* Auto-detect',
        'other_codec': '',
    }


def test__post_data_common_container(ptp_tracker_jobs, mocker):
    assert ptp_tracker_jobs._post_data_common_container == {
        'container': '* Auto-detect',
        'other_container': ''
    }


def test__post_data_common_resolution(ptp_tracker_jobs, mocker):
    assert ptp_tracker_jobs._post_data_common_resolution == {
        'resolution': '* Auto-detect',
        'other_resolution': ''
    }


@pytest.mark.parametrize(
    argnames='subtitles, ptp_subtitles, exp_post_data',
    argvalues=(
        pytest.param(
            [],
            {
                'No Subtitles': 0,
                'la-ta (forced)': 1,
                'la-ta': 2,
                'la (forced)': 3,
                'la': 4,
            },
            {'subtitles[]': [0]},
            id='No subtitles'
        ),
        pytest.param(
            [
                # Subtitles that have PTP codes
                Mock(language='la', territory='ta', forced=True),
                Mock(language='la', territory='ta', forced=False),
                Mock(language='la', territory='', forced=True),
                Mock(language='la', territory='', forced=False),

                # Subtitles that do not have PTP codes
                Mock(language='lx', territory='ta', forced=True),
                Mock(language='lx', territory='ta', forced=False),
                Mock(language='lx', territory='', forced=True),
                Mock(language='lx', territory='', forced=False),
            ],
            {
                'No Subtitles': 0,
                'la-ta (forced)': 1,
                'la-ta': 2,
                'la (forced)': 3,
                'la': 4,
            },
            {'subtitles[]': [1, 2, 3, 4]},
            id='Supported and unsupported subtitles'
        ),
        pytest.param(
            [
                # Subtitles that do not have PTP codes
                Mock(language='lx', territory='ta', forced=True),
                Mock(language='lx', territory='ta', forced=False),
                Mock(language='lx', territory='', forced=True),
                Mock(language='lx', territory='', forced=False),
            ],
            {
                'No Subtitles': 0,
                'la-ta (forced)': 1,
                'la-ta': 2,
                'la (forced)': 3,
                'la': 4,
            },
            {'subtitles[]': []},
            id='Only unsupported subtitles'
        ),
    ),
    ids=lambda v: str(v),
)
def test__post_data_common_subtitles(subtitles, ptp_subtitles, exp_post_data, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'subtitles', PropertyMock(return_value=subtitles))
    mocker.patch.dict(ptp.metadata.ptp_subtitles, ptp_subtitles, clear=True)
    assert ptp_tracker_jobs._post_data_common_subtitles == exp_post_data


def test__post_data_common_remaster(ptp_tracker_jobs, mocker):
    assert ptp_tracker_jobs._post_data_common_remaster == {}


@pytest.mark.parametrize(
    argnames='trumpable_job_output, exp_post_data',
    argvalues=(
        ((), {'trumpable[]': []}),
        (
            (
                str(ptp.metadata.PtpTrumpableReason.HARDCODED_SUBTITLES),
            ),
            {
                'trumpable[]': [
                    ptp.metadata.PtpTrumpableReason.HARDCODED_SUBTITLES.value,
                ],
            },
        ),
        (
            (
                str(ptp.metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES),
            ),
            {
                'trumpable[]': [
                    ptp.metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES.value,
                ],
            },
        ),
        (
            (
                str(ptp.metadata.PtpTrumpableReason.HARDCODED_SUBTITLES),
                str(ptp.metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES),
            ),
            {
                'trumpable[]': [
                    ptp.metadata.PtpTrumpableReason.HARDCODED_SUBTITLES.value,
                    ptp.metadata.PtpTrumpableReason.NO_ENGLISH_SUBTITLES.value,
                ],
            },
        ),
    ),
    ids=lambda v: str(v),
)
def test__post_data_common_trumpable(trumpable_job_output, exp_post_data, ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'trumpable_job', PropertyMock(return_value=Mock(
        output=trumpable_job_output,
    )))
    assert ptp_tracker_jobs._post_data_common_trumpable == exp_post_data


def test__post_data_add_movie_format(ptp_tracker_jobs, mocker):
    mocker.patch.object(type(ptp_tracker_jobs), 'ptp_group_id', PropertyMock(return_value='123456'))
    assert ptp_tracker_jobs._post_data_add_movie_format == {
        'groupid': ptp_tracker_jobs.ptp_group_id,
    }


def test__post_data_add_new_movie(ptp_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(ptp_tracker_jobs)

    mocker.patch.object(type(ptp_tracker_jobs), 'imdb_id', PropertyMock(return_value='tt123456'))

    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs.tracker, 'normalize_imdb_id', side_effect=lambda id: f'<{id}>'),
        'normalize_imdb_id',
    )
    mocks.attach_mock(
        mocker.patch.object(ptp_tracker_jobs, 'get_job_output', side_effect=(
            '<title>',
            '<year>',
            '<plot>',
            '<tags>',
            '<cover art>',
        )),
        'get_job_output',
    )

    exp_post_data = {
        'imdb': f'<{ptp_tracker_jobs.imdb_id}>',
        'title': '<title>',
        'year': '<year>',
        'album_desc': '<plot>',
        'tags': '<tags>',
        'image': '<cover art>',
    }
    assert ptp_tracker_jobs._post_data_add_new_movie == exp_post_data
    assert mocks.mock_calls == [
        call.normalize_imdb_id(ptp_tracker_jobs.imdb_id),
        call.get_job_output(ptp_tracker_jobs.title_job, slice=0),
        call.get_job_output(ptp_tracker_jobs.year_job, slice=0),
        call.get_job_output(ptp_tracker_jobs.plot_job, slice=0),
        call.get_job_output(ptp_tracker_jobs.tags_job, slice=0),
        call.get_job_output(ptp_tracker_jobs.cover_art_job, slice=0),
    ]



from unittest.mock import call, patch

import pytest

from upsies import errors
from upsies.utils import subproc


@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
def test_run_returns_stdout(run_mock):
    run_mock.return_value.stderr = ''
    run_mock.return_value.stdout = 'process output'
    stdout = subproc.run(['foo', 'bar', '--baz'])
    assert stdout == 'process output'
    assert run_mock.call_args_list == [call(
        ('foo', 'bar', '--baz'),
        shell=False,
        encoding='utf-8',
        stdout='Mocked PIPE',
        stderr='Mocked PIPE',
        stdin='Mocked PIPE',
    )]

@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
def test_run_raises_DependencyError_if_command_cannot_be_executed(run_mock):
    run_mock.side_effect = OSError()
    with pytest.raises(errors.DependencyError, match=r'^Missing dependency: foo$'):
        subproc.run(['foo', 'bar', '--baz'])

@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
def test_run_raises_ProcessError_if_stderr_is_truthy(run_mock):
    run_mock.return_value.stderr = 'something went wrong'
    run_mock.return_value.stdout = 'process output'
    with pytest.raises(errors.ProcessError, match=r'^something went wrong$'):
        subproc.run(('foo', 'bar', '--baz'))

@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
def test_run_ignores_stderr_on_request(run_mock):
    run_mock.return_value.stderr = 'something went wrong'
    run_mock.return_value.stdout = 'process output'
    stdout = subproc.run(('foo', 'bar', '--baz'), ignore_errors=True)
    assert stdout == 'process output'

@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
def test_run_caches_stdout_on_request(run_mock):
    run_mock.return_value.stderr = ''
    run_mock.return_value.stdout = 'process output'
    for _ in range(5):
        assert subproc.run(['foo', 'bar', '--baz'], cache=True) == 'process output'
        assert run_mock.call_args_list == [call(
            ('foo', 'bar', '--baz'),
            shell=False,
            encoding='utf-8',
            stdout='Mocked PIPE',
            stderr='Mocked PIPE',
            stdin='Mocked PIPE',
        )]

@patch('subprocess.run')
@patch('subprocess.PIPE', 'Mocked PIPE')
@patch('subprocess.STDOUT', 'Mocked STDOUT')
def test_run_joins_stdout_and_stderr_on_request(run_mock):
    run_mock.return_value.stdout = 'foo'
    run_mock.return_value.stderr = ''
    subproc.run(['foo'], join_stderr=True)
    assert run_mock.call_args_list == [call(
        ('foo',),
        shell=False,
        encoding='utf-8',
        stdout='Mocked PIPE',
        stderr='Mocked STDOUT',
        stdin='Mocked PIPE',
    )]



from unittest.mock import Mock, call

import pytest

from upsies.utils import predbs


def test_predbs(mocker):
    existing_predbs = (Mock(), Mock(), Mock())
    submodules_mock = mocker.patch('upsies.utils.submodules')
    subclasses_mock = mocker.patch('upsies.utils.subclasses', return_value=existing_predbs)
    assert predbs.predbs() == existing_predbs
    assert submodules_mock.call_args_list == [call('upsies.utils.predbs')]
    assert subclasses_mock.call_args_list == [call(predbs.PredbApiBase, submodules_mock.return_value)]


def test_predb_returns_ClientApiBase_instance(mocker):
    existing_predbs = (Mock(), Mock(), Mock())
    existing_predbs[0].configure_mock(name='foo')
    existing_predbs[1].configure_mock(name='bar')
    existing_predbs[2].configure_mock(name='baz')
    mocker.patch('upsies.utils.predbs.predbs', return_value=existing_predbs)
    assert predbs.predb('bar', config={'foo': 'bar'}) is existing_predbs[1].return_value
    assert existing_predbs[1].call_args_list == [call(config={'foo': 'bar'})]

def test_predb_fails_to_find_predb(mocker):
    existing_predbs = (Mock(), Mock(), Mock())
    existing_predbs[0].configure_mock(name='foo')
    existing_predbs[1].configure_mock(name='bar')
    existing_predbs[2].configure_mock(name='baz')
    mocker.patch('upsies.utils.predbs.predbs', return_value=existing_predbs)
    with pytest.raises(ValueError, match='^Unsupported scene release database: bam$'):
        predbs.predb('bam')
    for c in existing_predbs:
        assert c.call_args_list == []


def test_predb_names(mocker):
    existing_predbs = (Mock(), Mock(), Mock())
    existing_predbs[0].configure_mock(name='FOO')
    existing_predbs[1].configure_mock(name='bar')
    existing_predbs[2].configure_mock(name='Baz')
    mocker.patch('upsies.utils.predbs.predbs', return_value=existing_predbs)
    assert predbs.predb_names() == ['bar', 'Baz', 'FOO']
